

import os, sys, re, json


from .. import WSGIApplication, FailPage, GoTo, ValidateError, ServerError, use_submit_list, set_debug, skilift
from ..skilift.fromjson import get_defaults_from_file

from . import skiadminpackages

PROJECTFILES = os.path.dirname(os.path.realpath(__file__))
PROJECT = 'skiadmin'

# a search for anything none-alphanumeric and not an underscore
_AN = re.compile('[^\w]')



def start_call(called_ident, skicall):
    "Checks initial incoming call parameters, and using ident_data, retrieves session data and populates call_data"

    # initially populate call_data with some project info
    editedprojname = skicall.proj_data['editedprojname']

    projinfo = skilift.project_info(editedprojname)

    skicall.call_data = {'editedprojname':editedprojname,
                         'editedprojurl':projinfo.path,
                         'editedprojversion':projinfo.version,
                         'editedprojbrief':projinfo.brief,
                         'adminproj':skicall.project,
                         'extend_nav_buttons':[],
                         'caller_ident':skicall.caller_ident}

    if called_ident is None:
        # The call is to a url not found
        return

    # If caller_ident is not given there should be no further session data
    if not skicall.caller_ident:
        return called_ident

    # session data is received from the client as a json dictionary via ident_data
    if skicall.ident_data:
        session_data = json.loads(skicall.ident_data)
        # update call_data with session_data
        skicall.call_data.update(session_data)

    return called_ident


# skicall.submit_list defines packages, module, function to call
@use_submit_list
def submit_data(skicall):
    """The decorator calls the appropriate submit_data function, using skicall.submit_list
       if skicall.submit_list is invalid, then this function raises ServerError"""
    raise ServerError("skicall.submit_list invalid for responder %s,%s" % skicall.ident_list[-1])


def end_call(page_ident, page_type, skicall):
    """Sets navigation menus, and sends session data as ident_data."""

    # do not include session data or navigation for these types of pages
    if page_type in ('FilePage', 'CSS'):
        return

    call_data = skicall.call_data
    page_data = skicall.page_data

    editedprojname = call_data['editedprojname']

    projname, identnum = page_ident

    # do not include session data if target page is in another project
    if projname != PROJECT:
        return

    # do not include session or navigation if this is the responder map page
    if identnum == 26010:
        return

    if page_type == "TemplatePage":
        # header information
        if ("adminhead","page_head","large_text") not in page_data:
            page_data["adminhead","page_head","large_text"] = "Project: %s version: %s" % (editedprojname, call_data['editedprojversion'])
        # fill in navigation information
        set_navigation(identnum, call_data, page_data)

    # Show the status message
    if 'status' in call_data:
        page_data['foot', 'foot_status','footer_text'] = call_data['status']
        page_data[("adminhead","show_status","para_text")] = call_data['status']
        page_data[("adminhead","show_status","hide")] = False


    # get any required session data which has been set into call_data
    # and pass it to the client as ident_data

    sent_session_data = {}

    # this is a list of items to store in session data if they are available in call_data
    session_keys = ['location', 'field_arg', 'validx', 'module', 'widgetclass', 'widget_name', 'container', 'add_to_foldernumber', 
                    'section_name', 'schange', 'page_number', 'pchange', 'folder_number', 'fchange']

    for key, val in call_data.items():
        if key in session_keys:
            sent_session_data[key] = val

    if not sent_session_data:
        return

    # and this dictionary is inserted into ident_data as a json string
    page_data['ident_data'] = json.dumps(sent_session_data)


# As this project is not intended to run as a stand-alone service, a function
# is provided rather than an application object being immediately created.


def makeapp(editedprojname):
    """This function returns the skis application."""

    if not editedprojname:
        raise ServerError("The project name being edited is required")

    # get pallet of colours from defaults.json and place in proj_data
    # these will be used to populate w3-theme-ski.css and the Colours page

    adminbackcol = get_defaults_from_file(PROJECTFILES, PROJECT, key="backcol")
    if not adminbackcol:
        adminbackcol = "#bfb786"
    colours = get_defaults_from_file(PROJECTFILES, PROJECT, key="colours")
    if not colours:
        adminbackcol_rgb = skiadminpackages.css_styles.hex_int(adminbackcol)
        colours = skiadminpackages.css_styles.get_colours(*adminbackcol_rgb)


    proj_data = {"editedprojname":editedprojname, "colours":colours, "adminbackcol":adminbackcol}

    # The WSGIApplication created here is generally given a URL of "/skiadmin"
    # when added to the root project using application.add_project

    return WSGIApplication(project=PROJECT,
                           projectfiles=PROJECTFILES,
                           proj_data=proj_data,
                           start_call=start_call,
                           submit_data=submit_data,
                           end_call=end_call)


def set_navigation(identnum, call_data, page_data):
    "Sets the navigation buttons"

    # top navigation

    page_data["adminhead","top_nav","nav_links"] = [ [3, "Root Folder", False, ''] ]

    editedprojname = call_data['editedprojname']

    item_number = None

    if 'page_number' in call_data:
        item_number = call_data['page_number']
    elif 'folder_number' in call_data:
        item_number = call_data['folder_number']

    if item_number:
        # get list of [(name,number),...] of parents
        parents = skilift.parent_list(editedprojname, item_number)
        for item in parents:
            # for each item apart from root
            if item[1]:
                edited_folder = editedprojname + '_' + str(item[1])
                page_data["adminhead","top_nav","nav_links"].append(['edit_from_top_nav', item[0], False, edited_folder])
    # Top navigation lists in reverse order as the float right keeps making the next link the rightmost
    page_data["adminhead","top_nav","nav_links"].reverse()


    # Left navigation

    # uses a headers.NavButtons1 widget, each inner list consists of

    # 0 : The url, label or ident of the target page of the link
    # 1 : The displayed text of the link
    # 2 : If True, ident_data is sent with the link even if there is no get field data
    # 3 : The get field data to send with the link 

    if identnum == 20001:
        # main site index page
        page_data["left_nav","navbuttons","nav_links"] = [  [call_data['editedprojurl'], "Project", False, ''],
                                                            [3, "Root Folder", False, ''],
                                                            ["manage_specialpages", "Page Labels", False, ''],
                                                            ["manage_textblocks", "TextBlocks", False, ''],
                                                            ["manage_sections", "Sections", False, ''],
                                                            [300, "Colours", False, ''],
                                                            ['defaults', "Defaults", False, ''],
                                                            ["about_code", "Getting Started", False, '']
                                                           ]
    elif identnum == 510:
        # about_skis page
        page_data["left_nav","navbuttons","nav_links"] = [  [call_data['editedprojurl'], "Project", False, ''],
                                                            ['admin_home', "Admin", False, ''],
                                                            [3, "Root Folder", False, ''],
                                                            ["manage_specialpages", "Page Labels", False, ''],
                                                            ["manage_textblocks", "TextBlocks", False, ''],
                                                            ["manage_sections", "Sections", False, ''],
                                                            [300, "Colours", False, ''],
                                                            ['defaults', "Defaults", False, ''],
                                                            ["about_code", "Getting Started", False, '']
                                                           ]
    elif identnum == 85001:
        # operations index page
        page_data["left_nav","navbuttons","nav_links"] = [  [call_data['editedprojurl'], "Project", False, ''],
                                                            ['admin_home', "Admin", False, ''],
                                                            [3, "Root Folder", False, ''],
                                                            ["manage_specialpages", "Page Labels", False, ''],
                                                            ["manage_textblocks", "TextBlocks", False, ''],
                                                            ["manage_sections", "Sections", False, ''],
                                                            [300, "Colours", False, ''],
                                                            ["about_code", "Getting Started", False, '']
                                                           ]
    elif identnum == 20300:
        # colours index page
        page_data["left_nav","navbuttons","nav_links"] = [  [call_data['editedprojurl'], "Project", False, ''],
                                                            ['admin_home', "Admin", False, ''],
                                                            [3, "Root Folder", False, ''],
                                                            ["manage_specialpages", "Page Labels", False, ''],
                                                            ["manage_textblocks", "TextBlocks", False, ''],
                                                            ["manage_sections", "Sections", False, ''],
                                                            ['defaults', "Defaults", False, ''],
                                                            ["about_code", "Getting Started", False, '']
                                                           ]

    elif identnum == 70101:
        # about code page
        page_data["left_nav","navbuttons","nav_links"] = [  [call_data['editedprojurl'], "Project", False, ''],
                                                            ['admin_home', "Admin", False, ''],
                                                            [3, "Root Folder", False, ''],
                                                            ["manage_specialpages", "Page Labels", False, ''],
                                                            ["manage_textblocks", "TextBlocks", False, ''],
                                                            ["manage_sections", "Sections", False, ''],
                                                            [300, "Colours", False, ''],
                                                            ['defaults', "Defaults", False, '']
                                                           ]

    elif identnum == 25001:
        # page labels
        page_data["left_nav","navbuttons","nav_links"] = [  [call_data['editedprojurl'], "Project", False, ''],
                                                            ['admin_home', "Admin", False, ''],
                                                            [3, "Root Folder", False, ''],
                                                            ["manage_textblocks", "TextBlocks", False, ''],
                                                            ["manage_sections", "Sections", False, ''],
                                                            ['defaults', "Defaults", False, '']
                                                           ]


    elif ("left_nav","navbuttons","nav_links") not in page_data:
        # all other page have full suite of navigation buttons
        page_data["left_nav","navbuttons","nav_links"] = [  [call_data['editedprojurl'], "Project", False, ''],
                                                            ['admin_home', "Admin", False, ''],
                                                            [3, "Root Folder", False, ''],
                                                            ["manage_specialpages", "Page Labels", False, ''],
                                                            ["manage_textblocks", "TextBlocks", False, ''],
                                                            ["manage_sections", "Sections", False, '']
                                                           ]

    widget_info = None

    if item_number:
        item_info = skilift.item_info(editedprojname, item_number)
        if item_info:
            if (item_info.item_type == 'TemplatePage') or (item_info.item_type == 'SVG'):
                # add page name to navbuttons
                if item_info.item_type == 'TemplatePage':
                    page_data["left_nav","navbuttons","nav_links"].append(['back_to_page', item_info.name, True, ''])
                    page_data["left_nav","navbuttons","nav_links"].append(['page_head', item_info.name + ' - Head', True, ''])
                    page_data["left_nav","navbuttons","nav_links"].append(['page_body', item_info.name + ' - Body', True, ''])
                if item_info.item_type == 'SVG':
                    page_data["left_nav","navbuttons","nav_links"].append(['back_to_svgpage', item_info.name, True, ''])
                    page_data["left_nav","navbuttons","nav_links"].append(['page_svg', 'SVG', True, ''])
                if ('widget_name' in call_data) and ((item_info.item_type == 'TemplatePage') or (item_info.item_type == 'SVG')):
                    widget_info = skilift.widget_info(editedprojname, item_number, None, call_data['widget_name'])

    elif 'section_name' in call_data:
        page_data["left_nav","navbuttons","nav_links"].append(['back_to_section', call_data['section_name'], True, ''])
        if 'widget_name' in call_data:
            widget_info = skilift.widget_info(editedprojname, None, call_data['section_name'], call_data['widget_name'])

    if widget_info:
        # if widget has parent, display parent links
        display_parent(widget_info, page_data)
        page_data["left_nav","navbuttons","nav_links"].append(['retrieve_widget', widget_info.name, True, widget_info.name])
        # if widget has containers, display links to them
        if widget_info.containers:
            for cont in range(widget_info.containers):
                str_cont = str(cont)
                page_data["left_nav","navbuttons","nav_links"].append(['retrieve_container', widget_info.name + " " + str_cont, True, widget_info.name + "-" + str_cont])

    # add further buttons which may be set in call_data['extend_nav_buttons']
    if 'extend_nav_buttons' in call_data:
        if call_data['extend_nav_buttons']:
            page_data["left_nav","navbuttons","nav_links"].extend(call_data['extend_nav_buttons'])


def display_parent(widget_info, page_data):
    "Appends link to parent widget in navigation buttons"
    location = widget_info.location
    # location is (parent, container, location integers) where parent is section_name, head/body or parent widget name
    # if container is None then there is no parent widget
    if location[1] is None:
        return
    # widget is in a container
    # does the parent have parents?
    parent_name = location[0]
    parent_info = skilift.widget_info(widget_info.project, widget_info.pagenumber, widget_info.section_name, parent_name)
    # recursively display grandparents
    display_parent(parent_info, page_data)
    # display links to the parent widget
    page_data["left_nav","navbuttons","nav_links"].append(['retrieve_widget', parent_name, True, parent_name])
    page_data["left_nav","navbuttons","nav_links"].append(['retrieve_container', parent_name + " " + str(location[1]), True, parent_name + "-" + str(location[1])])







import os

# from skipole import the WSGIApplication class which will be used to create a wsgi
# application, and exception classes which you will need in your functions

from skipole import WSGIApplication, FailPage, GoTo, ValidateError, ServerError, use_submit_list

# the framework needs to know the location of the projectfiles directory holding the project data
# and static files.

PROJECTFILES = os.path.dirname(os.path.realpath(__file__))
PROJECT = "newproj"

# This file newproj.py is initially created under the projectfiles directory, however it does not
# need to be sited there, and can be moved and your code developed elsewhere. The above PROJECTFILES
# path is only needed to set the location of the support files used by the project, not code. 

# proj_data is an optional dictionary of values which you are free to use, it will be made
# available to your functions as the attribute 'skicall.proj_data'

PROJ_DATA={}


# Your code needs to provide your own version of the following three functions which will
# be used to set values into the widgets of returned pages. Minimal versions are provided below.


def start_call(called_ident, skicall):
    "When a call is initially received this function is called."
    return called_ident


# You may wish to apply the decorator '@use_submit_list' to the submit_data
# function below. See the skipole documentation for details.

def submit_data(skicall):
    "This function is called when a Responder wishes to submit data for processing in some manner"
    return


def end_call(page_ident, page_type, skicall):
    """This function is called at the end of a call prior to filling the returned page with skicall.page_data,
       it can also return an optional session cookie string."""
    return


# The above functions are required as arguments to the skipole.WSGIApplication object
# and will be called as required.

# create the wsgi application
application = WSGIApplication(project=PROJECT,
                              projectfiles=PROJECTFILES,
                              proj_data=PROJ_DATA,
                              start_call=start_call,
                              submit_data=submit_data,
                              end_call=end_call,
                              url="/")


                 ### add the 'skis' sub project ###

# The 'skis' application serves javascript and the w3.css files required by
# the framework widgets.

# The skis package, contains the function makeapp() - which returns a
# WSGIApplication object which is then appended to your own project

from skipole import skis

skis_application = skis.makeapp()
application.add_project(skis_application, url='/lib')

# The add_project method of application, enables the added sub application
# to be served at a URL which should extend the URL of the main 'root' application.
# The above shows the main application served at "/" and the skis library
# project served at "/lib"


#############################################################################
#
# You should remove everything below here when deploying and serving your
# finished application. The following lines are used to serve the project
# locally and add the skiadmin project for development.

# Normally, when deploying on a web server, you would follow the servers
# own documentation which should describe how to load a wsgi application.
# for example, using gunicorn3 by command line

# gunicorn3 -w 4 newproj:application

# Where gunicorn3 is the python3 version of gunicorn


if __name__ == "__main__":


    ############### THESE LINES ADD THE SKIADMIN SUB-PROJECT FOR DEVELOPMENT #
    ################# AND SHOULD BE REMOVED WHEN YOU DEPLOY YOUR APPLICATION #


    from skipole import skiadmin, set_debug, skilift
    set_debug(True)
    skiadmin_application = skiadmin.makeapp(editedprojname=PROJECT)
    application.add_project(skiadmin_application, url='/skiadmin')

    # serve the application with the development server from skilift

    host = "127.0.0.1"
    port = 8000
    print("Serving %s on port %s. Call http://localhost:%s/skiadmin to edit." % (PROJECT, port, port))
    skilift.development_server(host, port, application)
 






"""Defines widgets of page headers,
typically with titles, logo images and navigational buttons."""

from .. import skiboot
from .. import tag
from . import Widget, FieldArg, FieldArgList, FieldArgTable, FieldArgDict


class NavButtons1(Widget):
    """Defines a div, with a normally hidden error paragraph followed by a list of navigation buttons."""

    error_location = (0,0,0)

    arg_descriptions = {
                        'nav_links':FieldArgTable(("url", "text", "boolean", "text"), valdt=True),    # target, text in box, force_ident, get data
                        'button_class':FieldArg("cssclass", ''),
                        'li_class':FieldArg("cssclass", ''),
                        'ul_class':FieldArg("cssclass", ''),
                        'error_class':FieldArg("cssclass", "")
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
         nav_links: A list of lists, each inner list describing a link, the name of this field is used in the
               widgfield for any data returned in the get field
        For each navigation link, the table row is
          0 : The url, label or ident of the target page of the link
          1 : The displayed text of the link
          2 : If True, ident is appended to link even if there is no get field
          3 : The get field data to send with the link
        button_class: The class of the button links - which provides the appearance via CSS
        li_class: The class of the li tag
        ul_class: The class of the ul tag
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self[0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0] = tag.Part(tag_name="p")
        self[0][0][0] = ''
        self[1] = tag.Part(tag_name="ul")


    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the header"
        button_class = self.get_field_value('button_class')
        li_class = self.get_field_value('li_class')

        if self.get_field_value('error_class'):
            self[0].update_attribs({"class":self.get_field_value('error_class')})
        if self.error_status:
            self[0].del_one_attrib("style")
        if self.get_field_value('ul_class'):
            self[1].update_attribs({"class":self.get_field_value('ul_class')})

        # for each link in the nav_links table - create a link and add it
        # as a list item
        for row in self.get_field_value('nav_links'):
            linkurl, linktext, link_force_ident, link_getdata = row
            if not (linkurl or linktext):
                continue
            url = skiboot.get_url(linkurl, proj_ident=page.proj_ident)
            if not url:
                continue
            lnk = tag.Part(tag_name="a", text=linktext, attribs={"role":"button"})
            if button_class:
                lnk.update_attribs({"class":button_class})
            # create a url for the href
            get_field = {self.get_formname("nav_links"):link_getdata}
            url = self.make_get_url(page, url, get_field, link_force_ident)
            lnk.update_attribs({"href": url})
            if li_class:
                listtag = tag.Part(tag_name="li", attribs={'class':li_class})
            else:
                listtag = tag.Part(tag_name="li")
            listtag[0] = lnk
            self[1].append(listtag)

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
  <div>  <!-- normally hidden, class set to error_class -->
    <p> <!-- error message appears in this paragraph --> </p>
  </div>
  <ul>   <!-- class set to ul_class -->
    <li> <!-- class set to li_class -->
      <a role=\"button\" href=\"#\">  <!-- with class set to button_class -->
         <!-- The displayed text of the link -->
      </a>
    </li>
    <!-- further links -->
  </ul>
</div>"""




class NavButtons2(Widget):
    """Defines a div, containing navigation button links."""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {
                        'nav_links':FieldArgTable(("url", "text", "boolean", "text"), valdt=True),    # target, text in box, force_ident, get data
                        'button_class':FieldArg("cssclass", ''),
                        'button_style':FieldArg("cssstyle", '')
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
         nav_links: A list of lists, each inner list describing a link, the name of this field is used in the
               widgfield for any data returned in the get field
        For each navigation link, the table row is
          0 : The url, label or ident of the target page of the link
          1 : The displayed text of the link
          2 : If True, ident is appended to link even if there is no get field
          3 : The get field data to send with the link
        button_class: The class of the button links - which provides the appearance via CSS
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        # as this widget does not display errors, and is not json settable, hide if empty
        self.hide_if_empty=True


    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the list of links"
        button_class = self.get_field_value('button_class')
        button_style = self.get_field_value('button_style')

        # for each link in the nav_links table - create a link and add it
        # as a list item
        for row in self.get_field_value('nav_links'):
            linkurl, linktext, link_force_ident, link_getdata = row
            if not (linkurl or linktext):
                continue
            url = skiboot.get_url(linkurl, proj_ident=page.proj_ident)
            if not url:
                continue
            lnk = tag.Part(tag_name="a", text=linktext, attribs={"role":"button"})
            if button_class:
                lnk.update_attribs({"class":button_class})
            if button_style:
                lnk.update_attribs({"style":button_style})
            # create a url for the href
            get_field = {self.get_formname("nav_links"):link_getdata}
            url = self.make_get_url(page, url, get_field, link_force_ident)
            lnk.update_attribs({"href": url})
            self.append(lnk)

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
  <a role=\"button\" href=\"#\">  <!-- with class set to button_class -->
    <!-- The displayed text of the link -->
  </a>
  <!-- further links -->
</div>"""



class NavButtons3(Widget):
    """Defines a div, containing navigation button links."""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {
                        'link_ident':FieldArg("url", 'no_javascript'),
                        'json_ident':FieldArg("url", ''),
                        'button_text':FieldArgList('text', jsonset=True),
                        'get_field1':FieldArgList('text', valdt=True, jsonset=True),
                        'get_field2':FieldArgList('text', valdt=True, jsonset=True),
                        'button_classes':FieldArgList('text', jsonset=True),
                        'button_style':FieldArg("cssstyle", '')
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        json_ident: The url, ident or label to link, expecting a json file to be returned
        link_ident: The link to an html page taken if the client does not have javascript
        button_text: List of text strings to be placed within the links, defines the number of links
        get_field1: list of 'get' strings for each link set in the target url
        get_field2: list of second 'get' strings for each link set in the target url
        button_classes: list of CSS classes, each list item set to button links
        button_style: A style set on every button
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self._jsonurl = ''


    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the list of links"
        button_classes = self.get_field_value('button_classes')
        button_style = self.get_field_value('button_style')
        button_text = self.get_field_value('button_text')
        get_field1 = self.get_field_value('get_field1')
        get_field2 = self.get_field_value('get_field2')

        if not button_text:
            return

        link_ident = self.get_field_value("link_ident")
        if not link_ident:
            link_ident = 'no_javascript'
        # get url
        linkurl = skiboot.get_url(link_ident, proj_ident=page.proj_ident)

        json_ident = self.get_field_value("json_ident")
        if json_ident:
            self._jsonurl = skiboot.get_url(json_ident, proj_ident=page.proj_ident)

        # for each link - create a link and add it
        for index,txt in enumerate(button_text):
            lnk = tag.Part(tag_name="a", text=txt, attribs={"role":"button"})
            if button_style:
                lnk.update_attribs({"style":button_style})
            if button_classes:
                button_class = button_classes[index] if index < len(button_classes) else ''
                if button_class:
                    lnk.update_attribs({"class":button_class})

            get1 = get_field1[index] if index < len(get_field1) else ''
            get2 = get_field2[index] if index < len(get_field2) else ''
            get_fields = {self.get_formname("get_field1"):get1, self.get_formname("get_field2"):get2}
            url = self.make_get_url(page, linkurl, get_fields, True)
            lnk.update_attribs({"href": url})
            self.append(lnk)

    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets a click event handler"""
        jscript = """  $("#{ident} a").click(function (e) {{
    SKIPOLE.widgets['{ident}'].eventfunc(e);
    }});
""".format(ident = self.get_id())
        if self._jsonurl:
            return jscript + self._make_fieldvalues(jsonurl=self._jsonurl)


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
  <a role=\"button\" href=\"#\">  <!-- with class set to corresponding class in button_classes -->
    <!-- and get fields set to corresponding element in get_field1 and get_field2 -->
    <!-- with displayed button text set to corresponding text in button_text -->
  </a>
  <!-- further links -->
</div>"""








class TabButtons1(Widget):
    """Defines a div, containing tab buttons which can hide/display portions of the page.
       Hides all elements of the page with a given class, then displays a portion of the
       page with a given id"""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {
                        'tabs':FieldArgTable(("text", "text")), # text on button, id to display
                        'hide_class':FieldArg("cssclass", ''),
                        'button_class':FieldArg("cssclass", ''),
                        'button_style':FieldArg("cssstyle", ''),
                        'onclick_addclass':FieldArg("cssstyle", ''),
                        'onclick_removeclass':FieldArg("cssstyle", ''),
                        'active_button':FieldArg("integer", 0, jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        tabs: A list of lists, each inner list describing a button
          0 : The text to appear on the button
          1 : The id of the portion of the page to make visible
        hide_class: CSS class applied to page elements normally hidden.
        button_class: The CSS class of the buttons
        button_style: The button style
        onclick_addclass: CSS class to add to a button when it is clicked
        onclick_removeclass: CSS class to remove from a button when it is clicked
        active_button: the button index (starting at 0) of the button which is currently active
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self.hide_if_empty=True
        self._display_id_list = []
        self._hide_class = ''
        self._onclick_addclass = ''
        self._onclick_removeclass = ''
        self._active = 0


    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the list of buttons"

        # get parameters to send to javascript
        self._hide_class = self.get_field_value('hide_class')
        self._onclick_addclass = self.get_field_value('onclick_addclass')
        self._onclick_removeclass = self.get_field_value('onclick_removeclass')

        button_class = self.get_field_value('button_class')
        button_style = self.get_field_value('button_style')
        # for each button in the tabs table - create a button and add it

        self._active = self.get_field_value('active_button')

        for index, btn in enumerate(self.get_field_value('tabs')):
            buttontext, displayid = btn
            if not (buttontext or displayid):
                continue
            btn = tag.Part(tag_name="button", text=buttontext)
            if index == self._active:
                # this button flagged as chosen
                if button_class and self._onclick_addclass:
                    if button_class == self._onclick_addclass:
                        btn.update_attribs({"class":button_class})
                    else:
                        btn.update_attribs({"class":button_class + " " + self._onclick_addclass})
                elif button_class:
                    btn.update_attribs({"class":button_class})
                elif self._onclick_addclass:
                    btn.update_attribs({"class":self._onclick_addclass})
            else:
                # give each button button_class and self._onclick_removeclass which will be removed when the button is clicked
                if button_class and self._onclick_removeclass:
                    if button_class == self._onclick_removeclass:
                        btn.update_attribs({"class":button_class})
                    else:
                        btn.update_attribs({"class":button_class + " " + self._onclick_removeclass})
                elif button_class:
                    btn.update_attribs({"class":button_class})
                elif self._onclick_removeclass:
                    btn.update_attribs({"class":self._onclick_removeclass})
            if button_style:
                btn.update_attribs({"style":button_style})
            self.append(btn)
            self._display_id_list.append(displayid)


    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets a click event handler"""
        jscript = """  $("#{ident} button").click(function (e) {{
    SKIPOLE.widgets['{ident}'].eventfunc(e);
    }});
""".format(ident = self.get_id())
        other_parameters = {}
        if self._display_id_list:
            other_parameters['display_id_list'] = self._display_id_list
        if self._hide_class:
            other_parameters['hide_class'] = self._hide_class
        if self._onclick_addclass:
            other_parameters['onclick_addclass'] = self._onclick_addclass
        if self._onclick_removeclass:
            other_parameters['onclick_removeclass'] = self._onclick_removeclass
        if other_parameters:
            fieldvalues = self._make_fieldvalues(**other_parameters)
        else:
            fieldvalues = ""
        return jscript + fieldvalues + """  SKIPOLE.widgets["{ident}"].setbutton({active});
""".format(ident = self.get_id(), active=self._active)


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
  <button>  <!-- with class set to button_class -->
    <!-- The displayed text of the button -->
  </button>
  <!-- further buttons -->
</div>"""



class DropDownButton1(Widget):
    """Defines a div, button and links, typically used for drop down links."""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {
                        'nav_links':FieldArgTable(("url", "text", "boolean", "text"), valdt=True),    # target, text in link, force_ident, get data
                        'button_class':FieldArg("cssclass", ''),
                        'button_style':FieldArg("cssstyle", ''),
                        'button_text':FieldArg("text", 'Links'),
                        'div_class':FieldArg("cssclass", ''),
                        'link_class':FieldArg("cssclass", '')
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
         nav_links: A list of lists, each inner list describing a link, the name of this field is used in the
               widgfield for any data returned in the get field
        For each link, the table row is
          0 : The url, label or ident of the target page of the link
          1 : The displayed text of the link
          2 : If True, ident is appended to link even if there is no get field
          3 : The get field data to send with the link
        button_class: The CSS class of the button
        button_style: The CSS style of the button
        button_text: The text on the button
        div_class: CSS class of inner div holding links
        link_class: CSS class of each link
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self.hide_if_empty=True


    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the list of links"
        button_class = self.get_field_value('button_class')
        button_style = self.get_field_value('button_style')
        div_class = self.get_field_value('div_class')
        link_class = self.get_field_value('link_class')

        self[0] = tag.Part(tag_name="button", text=self.get_field_value('button_text'))
        if button_class:
            self[0].update_attribs({"class":button_class})
        if button_style:
            self[0].update_attribs({"style":button_style})

        if div_class:
            self[1] =  tag.Part(tag_name="div", attribs={"class":div_class})
        else:
            self[1] =  tag.Part(tag_name="div")

        # for each link in the nav_links table - create a link and add it
        # as a list item
        for row in self.get_field_value('nav_links'):
            linkurl, linktext, link_force_ident, link_getdata = row
            if not (linkurl or linktext):
                continue
            url = skiboot.get_url(linkurl, proj_ident=page.proj_ident)
            if not url:
                continue
            lnk = tag.Part(tag_name="a", text=linktext)
            if link_class:
                lnk.update_attribs({"class":link_class})
            # create a url for the href
            get_field = {self.get_formname("nav_links"):link_getdata}
            url = self.make_get_url(page, url, get_field, link_force_ident)
            lnk.update_attribs({"href": url})
            self[1].append(lnk)

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
  <button> <!-- With button_class and button_style -->
     <!-- Contains button text -->
  </button>
  <div>  <!-- With CSS class div_class -->
    <a href=\"#\">  <!-- with class set to link_class -->
      <!-- The displayed text of the link -->
    </a>
    <!-- further links -->
  </div>
</div>"""


class HeaderErrorPara(Widget):
    """A paragraph containing error text - normally hidden, displayed on error"""

    error_location = 0

    arg_descriptions = {
                        'hide':FieldArg("boolean", True, jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        hide: If True, sets display: none; on the widget, can be set/unset via JSON file
                    on error the value is overridden and the widget is shown
        """
        Widget.__init__(self, name=name, tag_name="p", brief=brief, **field_args)
        self[0] = self.error_message

    def _build(self, page, ident_list, environ, call_data, lang):
        # Hides widget if no error and hide is True
        self.widget_hide(self.get_field_value("hide"))


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<p>  <!-- with widget id and class widget_class, normally hidden,
          shown when an error is displayed. -->
  <!-- error message appears in this paragraph -->
</p>
"""


class HeadText(Widget):
    """A tag, normally a h1 tag, which can be specified, containing text"""

     # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'tag':FieldArg("text", 'h1'),
                        'large_text':FieldArg("text", "", jsonset=True)}

    def __init__(self, name=None, brief='', **field_args):
        """
        large_text: The text appearing in the tag
        """
        # pass fields to Widget
        Widget.__init__(self, name=name, tag_name="h1", brief=brief, **field_args)
        self[0] = ''

    def _build(self, page, ident_list, environ, call_data, lang):
        self.tag_name = self.get_field_value('tag')
        self[0] = self.get_field_value("large_text")

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<h1>  <!-- default h1, tag can be specified and with widget id and class widget_class -->
    <!-- set with large_text -->
</h1>"""


class HeaderText1(Widget):
    """Defines a div with large text (h1) and small text (p)."""

    error_location = (2,0,0)

    arg_descriptions = {
                        'large_text':FieldArg("text", '', jsonset=True),
                        'small_text':FieldArg("text", '', jsonset=True),
                        'show_error':FieldArg("text", default="", valdt=False, jsonset=True),
                        'error_class':FieldArg("cssclass", ""),
                        'h_style':FieldArg("cssstyle", ""),
                        'p_style':FieldArg("cssstyle", ""),
                        'h_class':FieldArg("cssclass", ""),
                        'p_class':FieldArg("cssclass", "")
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        large_text: The large text at the top of the page
        small_text:  The smaller text, at the top of the page, but beneath the large text
        error_class: The class applied to the paragraph containing the error message on error.
        h_style: CSS style applied to the h tag
        p_style: CSS style applied to the p tag
        h_class: CSS class applied to the h tag
        p_class: CSS class applied to the p tag
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self[0] = tag.Part(tag_name="h1")
        self[1] = tag.Part(tag_name="p")
        # error div at 2
        self[2] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[2][0] = tag.Part(tag_name="p")
        self[2][0][0] = ''

    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the header"
        self[0].text=self.get_field_value("large_text")
        self[1].text=self.get_field_value("small_text")
        if self.get_field_value("h_class"):
            self[0].update_attribs({"class":self.get_field_value("h_class")})
        if self.get_field_value("p_class"):
            self[1].update_attribs({"class":self.get_field_value("p_class")})
        if self.get_field_value("h_style"):
            self[0].update_attribs({"style":self.get_field_value("h_style")})
        if self.get_field_value("p_style"):
            self[1].update_attribs({"style":self.get_field_value("p_style")})
        if self.get_field_value('error_class'):
            self[2].update_attribs({"class":self.get_field_value('error_class')})
        if self.error_status:
            self[2].del_one_attrib("style")



    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div> <!--  with widget id and class widget_class -->
  <h1> <!-- content set to large_text --> </h1>
  <p> <!-- content set to small_text --> </p>
  <div style="display:none;"> <!-- class set to error_class -->
    <p> <!-- content set to show_error --> </p>
  </div>
</div>"""


class HeaderText2(Widget):
    """Defines a div with large text (h2) and small text (p)."""

    error_location = (2,0,0)

    arg_descriptions = {
                        'large_text':FieldArg("text", '', jsonset=True),
                        'small_text':FieldArg("text", '', jsonset=True),
                        'show_error':FieldArg("text", default="", valdt=False, jsonset=True),
                        'error_class':FieldArg("cssclass", ""),
                        'h_style':FieldArg("cssstyle", ""),
                        'p_style':FieldArg("cssstyle", ""),
                        'h_class':FieldArg("cssclass", ""),
                        'p_class':FieldArg("cssclass", "")
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        large_text: The large text at the top of the page
        small_text:  The smaller text, at the top of the page, but beneath the large text
        error_class: The class applied to the paragraph containing the error message on error.
        h_style: CSS style applied to the h tag
        p_style: CSS style applied to the p tag
        h_class: CSS class applied to the h tag
        p_class: CSS class applied to the p tag
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self[0] = tag.Part(tag_name="h2")
        self[1] = tag.Part(tag_name="p")
        # error div at 2
        self[2] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[2][0] = tag.Part(tag_name="p")
        self[2][0][0] = ''


    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the header"
        self[0].text=self.get_field_value("large_text")
        self[1].text=self.get_field_value("small_text")
        if self.get_field_value("h_class"):
            self[0].update_attribs({"class":self.get_field_value("h_class")})
        if self.get_field_value("p_class"):
            self[1].update_attribs({"class":self.get_field_value("p_class")})
        if self.get_field_value("h_style"):
            self[0].update_attribs({"style":self.get_field_value("h_style")})
        if self.get_field_value("p_style"):
            self[1].update_attribs({"style":self.get_field_value("p_style")})
        if self.get_field_value('error_class'):
            self[2].update_attribs({"class":self.get_field_value('error_class')})
        if self.error_status:
            self[2].del_one_attrib("style")



    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div> <!--  with widget id and class widget_class -->
  <h2> <!-- content set to large_text --> </h2>
  <p> <!-- content set to small_text --> </p>
  <div style="display:none;"> <!-- class set to error_class -->
    <p> <!-- content set to show_error --> </p>
  </div>
</div>"""


class HeaderText3(Widget):
    """Defines a div with large text (h3) and small text (p)."""

    error_location = (2,0,0)

    arg_descriptions = {
                        'large_text':FieldArg("text", '', jsonset=True),
                        'small_text':FieldArg("text", '', jsonset=True),
                        'show_error':FieldArg("text", default="", valdt=False, jsonset=True),
                        'error_class':FieldArg("cssclass", ""),
                        'h_style':FieldArg("cssstyle", ""),
                        'p_style':FieldArg("cssstyle", ""),
                        'h_class':FieldArg("cssclass", ""),
                        'p_class':FieldArg("cssclass", "")
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        large_text: The large text at the top of the page
        small_text:  The smaller text, at the top of the page, but beneath the large text
        error_class: The class applied to the paragraph containing the error message on error.
        h_style: CSS style applied to the h tag
        p_style: CSS style applied to the p tag
        h_class: CSS class applied to the h tag
        p_class: CSS class applied to the p tag
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self[0] = tag.Part(tag_name="h3")
        self[1] = tag.Part(tag_name="p")
        # error div at 2
        self[2] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[2][0] = tag.Part(tag_name="p")
        self[2][0][0] = ''


    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the header"
        self[0].text=self.get_field_value("large_text")
        self[1].text=self.get_field_value("small_text")
        if self.get_field_value("h_class"):
            self[0].update_attribs({"class":self.get_field_value("h_class")})
        if self.get_field_value("p_class"):
            self[1].update_attribs({"class":self.get_field_value("p_class")})
        if self.get_field_value("h_style"):
            self[0].update_attribs({"style":self.get_field_value("h_style")})
        if self.get_field_value("p_style"):
            self[1].update_attribs({"style":self.get_field_value("p_style")})
        if self.get_field_value('error_class'):
            self[2].update_attribs({"class":self.get_field_value('error_class')})
        if self.error_status:
            self[2].del_one_attrib("style")



    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div> <!-- with widget id and class widget_class -->
  <h3> <!-- content set to large_text --> </h3>
  <p> <!-- content set to small_text --> </p>
  <div style="display:none;"> <!-- class set to error_class -->
    <p> <!-- content set to show_error --> </p>
  </div>
</div>"""


class HeaderText4(Widget):
    """Defines a div with large text (h4) and small text (p)."""

    error_location = (2,0,0)

    arg_descriptions = {
                        'large_text':FieldArg("text", '', jsonset=True),
                        'small_text':FieldArg("text", '', jsonset=True),
                        'show_error':FieldArg("text", default="", valdt=False, jsonset=True),
                        'error_class':FieldArg("cssclass", ""),
                        'h_style':FieldArg("cssstyle", ""),
                        'p_style':FieldArg("cssstyle", ""),
                        'h_class':FieldArg("cssclass", ""),
                        'p_class':FieldArg("cssclass", "")
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        large_text: The large text at the top of the page
        small_text:  The smaller text, at the top of the page, but beneath the large text
        error_class: The class applied to the paragraph containing the error message on error.
        h_style: CSS style applied to the h tag
        p_style: CSS style applied to the p tag
        h_class: CSS class applied to the h tag
        p_class: CSS class applied to the p tag
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self[0] = tag.Part(tag_name="h4")
        self[1] = tag.Part(tag_name="p")
        # error div at 2
        self[2] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[2][0] = tag.Part(tag_name="p")
        self[2][0][0] = ''


    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the header"
        self[0].text=self.get_field_value("large_text")
        self[1].text=self.get_field_value("small_text")
        if self.get_field_value("h_class"):
            self[0].update_attribs({"class":self.get_field_value("h_class")})
        if self.get_field_value("p_class"):
            self[1].update_attribs({"class":self.get_field_value("p_class")})
        if self.get_field_value("h_style"):
            self[0].update_attribs({"style":self.get_field_value("h_style")})
        if self.get_field_value("p_style"):
            self[1].update_attribs({"style":self.get_field_value("p_style")})
        if self.get_field_value('error_class'):
            self[2].update_attribs({"class":self.get_field_value('error_class')})
        if self.error_status:
            self[2].del_one_attrib("style")



    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div> <!-- with widget id and class widget_class -->
  <h4> <!-- content set to large_text --> </h4>
  <p> <!-- content set to small_text --> </p>
  <div style="display:none;"> <!-- class set to error_class -->
    <p> <!-- content set to show_error --> </p>
  </div>
</div>"""


class HeaderText5(Widget):
    """Defines a div with large text (h5) and small text (p)."""

    error_location = (2,0,0)

    arg_descriptions = {
                        'large_text':FieldArg("text", '', jsonset=True),
                        'small_text':FieldArg("text", '', jsonset=True),
                        'show_error':FieldArg("text", default="", valdt=False, jsonset=True),
                        'error_class':FieldArg("cssclass", ""),
                        'h_style':FieldArg("cssstyle", ""),
                        'p_style':FieldArg("cssstyle", ""),
                        'h_class':FieldArg("cssclass", ""),
                        'p_class':FieldArg("cssclass", "")
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        large_text: The large text at the top of the page
        small_text:  The smaller text, at the top of the page, but beneath the large text
        error_class: The class applied to the paragraph containing the error message on error.
        h_style: CSS style applied to the h tag
        p_style: CSS style applied to the p tag
        h_class: CSS class applied to the h tag
        p_class: CSS class applied to the p tag
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self[0] = tag.Part(tag_name="h5")
        self[1] = tag.Part(tag_name="p")
        # error div at 2
        self[2] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[2][0] = tag.Part(tag_name="p")
        self[2][0][0] = ''


    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the header"
        self[0].text=self.get_field_value("large_text")
        self[1].text=self.get_field_value("small_text")
        if self.get_field_value("h_class"):
            self[0].update_attribs({"class":self.get_field_value("h_class")})
        if self.get_field_value("p_class"):
            self[1].update_attribs({"class":self.get_field_value("p_class")})
        if self.get_field_value("h_style"):
            self[0].update_attribs({"style":self.get_field_value("h_style")})
        if self.get_field_value("p_style"):
            self[1].update_attribs({"style":self.get_field_value("p_style")})
        if self.get_field_value('error_class'):
            self[2].update_attribs({"class":self.get_field_value('error_class')})
        if self.error_status:
            self[2].del_one_attrib("style")



    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div> <!-- with widget id and class widget_class -->
  <h5> <!-- content set to large_text --> </h5>
  <p> <!-- content set to small_text --> </p>
  <div style="display:none;"> <!-- class set to error_class -->
    <p> <!-- content set to show_error --> </p>
  </div>
</div>"""


class HeaderText6(Widget):
    """Defines a div with large text (h6) and small text (p)."""

    error_location = (2,0,0)

    arg_descriptions = {
                        'large_text':FieldArg("text", '', jsonset=True),
                        'small_text':FieldArg("text", '', jsonset=True),
                        'show_error':FieldArg("text", default="", valdt=False, jsonset=True),
                        'error_class':FieldArg("cssclass", ""),
                        'h_style':FieldArg("cssstyle", ""),
                        'p_style':FieldArg("cssstyle", ""),
                        'h_class':FieldArg("cssclass", ""),
                        'p_class':FieldArg("cssclass", "")
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        large_text: The large text at the top of the page
        small_text:  The smaller text, at the top of the page, but beneath the large text
        error_class: The class applied to the paragraph containing the error message on error.
        h_style: CSS style applied to the h tag
        p_style: CSS style applied to the p tag
        h_class: CSS class applied to the h tag
        p_class: CSS class applied to the p tag
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self[0] = tag.Part(tag_name="h6")
        self[1] = tag.Part(tag_name="p")
        # error div at 2
        self[2] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[2][0] = tag.Part(tag_name="p")
        self[2][0][0] = ''


    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the header"
        self[0].text=self.get_field_value("large_text")
        self[1].text=self.get_field_value("small_text")
        if self.get_field_value("h_class"):
            self[0].update_attribs({"class":self.get_field_value("h_class")})
        if self.get_field_value("p_class"):
            self[1].update_attribs({"class":self.get_field_value("p_class")})
        if self.get_field_value("h_style"):
            self[0].update_attribs({"style":self.get_field_value("h_style")})
        if self.get_field_value("p_style"):
            self[1].update_attribs({"style":self.get_field_value("p_style")})
        if self.get_field_value('error_class'):
            self[2].update_attribs({"class":self.get_field_value('error_class')})
        if self.error_status:
            self[2].del_one_attrib("style")

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div> <!-- with widget id and class widget_class -->
  <h6> <!-- content set to large_text --> </h6>
  <p> <!-- content set to small_text --> </p>
  <div style="display:none;"> <!-- class set to error_class -->
    <p> <!-- content set to show_error --> </p>
  </div>
</div>"""








"""Functions for editing a folder"""

from ..ski import skiboot, read_json
from ..ski.excepts import ServerError

from . import project_loaded, item_info, ident_exists


def _get_folder(project, foldernumber):
    """On success return (deep copy of folder, empty string)
       on failure return (None, failure string)"""
    if not project_loaded(project, error_if_not=False):
        return None, "Given project is not loaded"
    if not isinstance(foldernumber, int):
        return None, "Given foldernumber is not an integer"
    ident = skiboot.Ident.to_ident((project, foldernumber))
    if ident is None:
        return None, "Invalid project, foldernumber"
    folder = skiboot.from_ident(ident, project)
    if folder is None:
        return None, "Invalid Folder"
    if folder.page_type != 'Folder':
        return None, "Item is not a Folder"
    return folder, ""


def folderchange(project, foldernumber):
    """Returns the folder change UUID string

       Given a project name, and the integer ident number of a folder within the project,
       returns a 'change' uuid associated  with the folder. A new random uuid is generated
       every time a folder is changed, and is therefore a measure that a given folder both
       exists, and has not changed since the last time it was accessed.
       The project must be currently loaded as either the root project or a sub-project.
       Returns None if no folder with this foldernumber is found."""
    # raise error if invalid project
    project_loaded(project)
    info = item_info(project, foldernumber)
    if not info:
        return
    if info.item_type != "Folder":
        return
    return info.change


def rename_folder(project, foldernumber, fchange, newname):
    """Rename the folder, return folder change uuid on success, raises ServerError on failure

       Given a project name, the integer ident number of a folder within the project,
       the folder change uuid and a new folder name, it changes the folder name and
       returns a new uuid on success, or raises a ServerError on failure.
       The project must be currently loaded as either the root project or a sub-project."""
    if not newname:
        raise ServerError(message="No new folder name given")
    # get a copy of the folder, which can have a new name set
    # and can then be saved to the project
    folder, error_message = _get_folder(project, foldernumber)
    if folder is None:
        raise ServerError(message=error_message)
    if folder.ident.num == 0:
        raise ServerError(message="Cannot rename the root folder")
    if folder.change != fchange:
        raise ServerError(message="The folder has been changed prior to this submission, someone else may be editing this project")
    proj = skiboot.getproject(project)
    if proj is None:
        raise ServerError(message="Project not loaded")
    # Rename the folder
    folder.name = newname
    # And save this folder copy to the project
    return proj.save_folder(folder)


def delete_folder(project, foldernumber):
    "delete this folder, return None on success, error message on failure"
    folder, error_message = _get_folder(project, foldernumber)
    if folder is None:
        return error_message
    editedproj = skiboot.getproject(project)
    if editedproj is None:
        return "Project not loaded"
    if folder.page_type != "Folder":
        return "Invalid folder"
    if folder.ident.num == 0:
        return "Cannot delete the root folder"
    if len(folder):
        return "Cannot delete a folder with contents"
    try:
        # delete the folder from the project
        editedproj.delete_item(folder.ident)
    except ServerError as e:
        return e.message


def delete_folder_recursively(project, foldernumber, fchange):
    "delete this folder, raise ServerError on failure, returns parentfolder number and change when done"
    folder, error_message = _get_folder(project, foldernumber)
    if folder is None:
        raise ServerError(message=error_message)
    editedproj = skiboot.getproject(project)
    if editedproj is None:
        raise ServerError(message="Project not loaded")
    if folder.page_type != "Folder":
        raise ServerError(message="Invalid folder")
    if folder.ident.num == 0:
        raise ServerError(message="Cannot delete the root folder")
    if folder.change != fchange:
        raise ServerError(message="The folder has been changed prior to this submission, someone else may be editing this project")
    # delete the folder from the project
    return editedproj.delete_folder_recursively(folder.ident)



def folder_description(project, foldernumber, fchange, brief):
    "Set a folder brief description, return folder change uuid on success, raises ServerError on failure"
    if not brief:
        return "No new folder name given"
    # get a copy of the folder, which can have a new brief set
    # and can then be saved to the project
    folder, error_message = _get_folder(project, foldernumber)
    if folder is None:
        return error_message
    if folder.change != fchange:
        raise ServerError(message="The folder has been changed prior to this submission, someone else may be editing this project")
    proj = skiboot.getproject(project)
    if proj is None:
        raise ServerError(message="Project not loaded")
    # Set the folder brief
    folder.brief = brief
    # save the altered folder, and return the folder.change uuid
    return proj.save_folder(folder)


def set_default_page(project, foldernumber, fchange, default_page_name):
    "set new default page on this folder, return folder change uuid on success, raises ServerError on failure"
    # get a copy of the folder
    # which can then be saved to the project
    folder, error_message = _get_folder(project, foldernumber)
    if folder is None:
        raise ServerError(message=error_message)
    if folder.change != fchange:
        raise ServerError(message="The folder has been changed prior to this submission, someone else may be editing this project")
    proj = skiboot.getproject(project)
    if proj is None:
        raise ServerError(message="Project not loaded")
    if not default_page_name:
        default_page_name = ''
    elif default_page_name not in folder.pages:
        raise ServerError(message="The page to set as default has not been found in this folder")
    folder.default_page_name = default_page_name
    # save the altered folder, and return the folder.change uuid
    return proj.save_folder(folder)


def set_restricted_status(project, foldernumber, fchange, restricted):
    """set restricted True, set this folder restricted, or unrestricted if False.
       r, return folder change uuid on success, raises ServerError on failure"""
    # get a copy of the folder
    # which can then be saved to the project
    if foldernumber == 0:
        return "Cannot change the root folder restricted status"
    folder, error_message = _get_folder(project, foldernumber)
    if folder is None:
        raise ServerError(message=error_message)
    if folder.change != fchange:
        raise ServerError(message="The folder has been changed prior to this submission, someone else may be editing this project")
    proj = skiboot.getproject(project)
    if proj is None:
        raise ServerError(message="Project not loaded")
    if restricted:
        # set folder and sub items as retricted
        restricted_list = folder.set_restricted()
        if not restricted_list:
            raise ServerError(message="No action taken")
        # And save this folder copy to the project
        for f in restricted_list:
            proj.save_folder(f)
        return folderchange(project, foldernumber)
    # unset restricted if parent is not restricted
    if folder.parentfolder.restricted:
        raise ServerError(message="Parent folder is restricted, cannot set this folder as unrestricted")
    # un-restrict the folder
    status = folder.set_unrestricted()
    if not status:
        raise ServerError(message="Failed to set unrestricted")
    # And save this folder copy to the project
    # save the altered folder, and return the folder.change uuid
    return proj.save_folder(folder)


def _check_parent_number(project, parent_number):
    "Checks parent_number is a Folder and returns parentinfo"
    project_loaded(project)
    if not isinstance(parent_number, int):
        raise ServerError(message="parent_number is not an integer")
    parentinfo = item_info(project, parent_number)
    if not parentinfo:
        raise ServerError(message = "Invalid parent folder")
    # parentinfo is a named tuple with members
    # 'project', 'project_version', 'itemnumber', 'item_type', 'name', 'brief', 'path', 'label_list', 'change', 'parentfolder_number', 'restricted'

    if parentinfo.item_type != "Folder":
        raise ServerError(message = "The parent with this ident number is not a folder")
    return parentinfo


def move_to_folder(project, item_number, new_folder_number):
    "Moves a given item - a folder or page, to a new parent folder"
    project_loaded(project)
    if not isinstance(item_number, int):
        raise ServerError(message="item_number is not an integer")
    if not isinstance(new_folder_number, int):
        raise ServerError(message="new parent folder number is not an integer")

    item = skiboot.from_ident(item_number, proj_ident=project)
    if not item:
        raise ServerError(message="No valid item to move given")

    folder = skiboot.from_ident(new_folder_number, proj_ident=project)
    if not folder:
        raise ServerError(message="No valid target folder given")

    if folder.page_type != 'Folder':
        raise ServerError(message="Target item is not a folder")

    old_parent = item.parentfolder
    if old_parent == folder:
       raise ServerError(message="Parent folder unchanged?")

    if item.name in folder:
        raise ServerError(message="The folder already contains an item with this name")

    editedproj = skiboot.getproject(project)
    editedproj.save_item(item, folder.ident)


def make_new_folder(project, parent_number, folder_dict):
    """Creates a new folder, raise ServerError on failure, returns new folder ident number
    folder_dict is something like:

    {
     "name":"folder_name",
     "ident":999,
     "brief":"brief description of the folder",
     "restricted":False
    }

    if ident not given, this function chooses the next free ident number.
    """
    parentinfo = _check_parent_number(project, parent_number)
    if 'ident' not in folder_dict:
        editedproj = skiboot.getproject(project)
        folder_dict['ident'] = editedproj.max_ident_num+1
    elif ident_exists(project, folder_dict['ident']):
        raise ServerError(message = "An item with this ident number already exists")

    # create the folder
    ident = read_json.create_folder(project, parent_number, 0, folder_dict["name"], parentinfo.restricted, folder_dict)
    return ident.num


def make_new_page(project, parent_number, page_dict):
    """Creates a new page, raise ServerError on failure, returns new page ident number
    page_dict is something like:

    {
    "name":"normalize.css",
    "ident": 1002,
    "brief": "From necolas.github.io/normalize.css/",
    "FilePage": {
        "filepath": "newproj/static/css/normalize.css",
        "enable_cache": True,
        "mimetype": "text/css"
        }
    }

    if ident not given, this function chooses the next free ident number.
    """
    parentinfo = _check_parent_number(project, parent_number)
    if 'ident' in page_dict:
        # check it is an integer
        if not isinstance(page_dict['ident'], int):
            raise ServerError(message = "The new page ident number must be an integer")
    else:
        # no ident given, get next free number
        editedproj = skiboot.getproject(project)
        page_dict['ident'] = editedproj.max_ident_num+1
    if ident_exists(project, page_dict['ident']):
        raise ServerError(message = "An item with this ident number already exists")
    # create the page
    if "SVG" in page_dict:
        read_json.create_svgpage(project, parent_number, page_dict['ident'], page_dict['name'], page_dict['brief'], page_dict)
    elif "TemplatePage" in page_dict:
        read_json.create_templatepage(project, parent_number, page_dict['ident'], page_dict['name'], page_dict['brief'], page_dict)
    elif "FilePage" in page_dict:
        read_json.create_filepage(project, parent_number, page_dict['ident'], page_dict['name'], page_dict['brief'], page_dict)
    elif "CSS" in page_dict:
        read_json.create_csspage(project, parent_number, page_dict['ident'], page_dict['name'], page_dict['brief'], page_dict)
    elif "JSON" in page_dict:
        read_json.create_jsonpage(project, parent_number, page_dict['ident'], page_dict['name'], page_dict['brief'], page_dict)
    elif "RespondPage" in page_dict:
        read_json.create_respondpage(project, parent_number, page_dict['ident'], page_dict['name'], page_dict['brief'], page_dict)
    else:
        raise ServerError("page data not recognized")
    return page_dict['ident']


def copy_page(project, pagenumber, foldernumber, new_page_number, new_name, brief):
    """Copy an existing page given by pagenumber, to a new page number with new name and brief into the folder given by foldernumber
       return the folder.change uuid"""
    if not isinstance(new_page_number, int):
        raise ServerError(message = "The new page ident number must be an integer")
    orig_page = skiboot.from_ident(pagenumber, project)
    if orig_page is None:
        raise ServerError(message = "The ident to be copied is not recognised")
    if orig_page.page_type == "Folder":
        raise ServerError("Invalid item to be copied")
    parentinfo = _check_parent_number(project, foldernumber)
    folder = skiboot.from_ident(parentinfo.itemnumber, project)
    # create new page
    new_page = orig_page.copy(name=new_name, brief = brief)
    # add page to parent folder
    folder.add_page(new_page, new_page_number)
    return folder.change







"""
This module defines the SkipoleProject class and SkiCall class

SkipoleProject being the core of the project which loads
the project from JSON files and responds to incoming calls by calling the user functions

SkiCall is the class of the skicall object which is created for each incoming
call and is passed as an argument to the user functions 
"""


import copy, os, cgi, collections, html, pprint, json, shutil, uuid, sys, traceback, re

from base64 import urlsafe_b64decode

from http import cookies

# a search for anything none-alphanumeric and not an underscore
_AN = re.compile('[^\w]')

# a search for anything none-alphanumeric, not an underscore, not -
# to allow any character in base64.urlsafe_b64encode/base64.urlsafe_b64decode
# note the = sign is not allowed, though in b64encode it is used for padding
# This is because padding is removed on sending, and added here on receiving

_AN64 = re.compile('[^\w\-]')

from . import skiboot, read_json
from .excepts import ValidateError, ServerError, FailPage, ErrorMessage, GoTo, PageError
from .. import textblocks


# ServerError raised in this module use codes 9000 to 9100


# These three are 'default' functions, used if the
# user functions are not given 

def _start_call(called_ident, skicall):
    "Default start_call function"
    return called_ident

def _submit_data(skicall):
    "Default submit_data function"
    return

def _end_call(page_ident, page_type, skicall):
    "Default end_call function"
    return



class SkipoleProject(object):
    """The SkipoleProject - an instance being a callable WSGI application"""

    def __init__(self, project, projectfiles, proj_data={}, start_call=None, submit_data=None, end_call=None, url="/"):
        """Loads the project from JSON files and records the user functions"""
        if _AN.search(project):
            raise ServerError(message="Error: Invalid project name, alphanumeric only")
        if '_' in project:
            raise ServerError(message="Error: Invalid project name, alphanumeric only (no underscore).")
        self._proj_ident = project
        self.projectfiles = projectfiles
        self.proj_data = proj_data
        if start_call is None:
            self.start_call = _start_call
        else:
            self.start_call = start_call
        if submit_data is None:
            self.submit_data = _submit_data
        else:
            self.submit_data = submit_data
        if end_call is None:
            self.end_call = _end_call
        else:
            self.end_call = end_call

        # initially, assume this is the root projcet, and sub projects can be added
        self.rootproject = True

        # initial values, will be set from the json file
        self.brief = "Project %s" % project
        self.version = "0.0.0"
        # The url of the root folder
        url=url.strip("/").lower()
        if url:
            self.url = "/" + url + "/"
        else:
            self.url = "/"
        # The root Folder
        self.root = None
        # dictionary of special pages, key = label: value = page ident
        self.special_pages = {}
        # dictionary of sections, key = name: value = section
        self.sections = {}

        # an ordered dictionary of {proj_ident: url,...}, ordered by length of url
        self._subproject_paths = collections.OrderedDict()
        # self.subprojects is a dictionary of sub projects {proj_ident: Project instance,.....}
        self.subprojects = {}

        # dictionary of idents: to folder or page, apart from root
        # note keys are full Ident instances, values are  folder or page instances
        self.identitems = {}

        # Create an instance of the AccessTextBlocks class for this project
        self.textblocks = textblocks.AccessTextBlocks(self._proj_ident, projectfiles, skiboot.default_language())

        # maintain a cach dictionary of paths against idents {path:ident}
        self._paths = {}

        # load project from json files
        self.load_from_json()

        # and add this project to the project register
        skiboot.add_to_project_register(self)


    def set_default_language(self, language):
        "Sets the project default language"
        self.textblocks.default_language = language

    def get_default_language(self):
        "Returns the project default language"
        return self.textblocks.default_language

    default_language = property(get_default_language, set_default_language)


    # The __call__ method makes the instance callable, as required by a wsgi application

    # This method first gets cookies from method self.get_cookies()
    # it then gets status, headers and data from method self.respond()

    # self.respond() parses the requested path, and calls self.proj_respond() if the page
    # requested is part of this project, or it calls subproj.proj_respond() if the page
    # requested is part of a sub project.
    # It returns a page not found if the path does not match any page.
    # It checks for a ServerError, and returns the server error page if one is raised.

    # self.proj_respond(), parses caller ident and called ident and calls
    # self.proj_start_call() which creates the skicall object and calls the users start_call function
    # self.proj_respond() checks the ident returned from self.proj_start_call(), it then calls
    # self.read_form_data() to read any any form data submitted in the raw data, and finally calls
    # self.status_headers_data() or subproj.status_headers_data() if the page returned from start_call
    # is a page of a subproject. It checks for a ValidateError, and returns the validate error page if one is raised.

    # self.status_headers_data() calls responders, and finally calls end_call, returning the wanted status, headers and data



    def __call__(self, environ, start_response):
        "Defines this projects callable as the wsgi application"
        # get received cookies, and lang which is a tuple of (preferred language, default language)
        lang, received_cookies = self.get_cookies(environ)
        status, headers, data = self.respond(environ, lang, received_cookies)
        start_response(status, headers)
        return data


    def get_cookies(self, environ):
        """Gets cookies from environ. and places them in 'received_cookies' dictionary.
           Checks presence of language in received cookie, or environ and creates a 'lang' tuple
           consisting of (preferred language, default language)
           returns lang, received_cookies"""
        try:
            cookiemorsals = cookies.SimpleCookie(environ["HTTP_COOKIE"])
        except Exception:
            cookiemorsals = None
        if cookiemorsals:
            received_cookies = {item:m.value for item,m in cookiemorsals.items()}
        else:
            received_cookies = {}
        if 'language' in received_cookies:
            language = received_cookies["language"]
        else:
            if "HTTP_ACCEPT_LANGUAGE" in environ:
                language_list = environ["HTTP_ACCEPT_LANGUAGE"].split(',')
                language = language_list[0]
            else:
                language = self.default_language
        lang = (language, self.default_language)
        return lang, received_cookies


    def respond(self, environ, lang, received_cookies):
        """After cookies obtained, this is called by __call__, and returns status, headers, data
           Finds the path called, and passes the path to self.proj_respond()
           or to subproj.proj_respond() if the path indicates the call is to a sub project.
           Detects if a ServerError occurs, and if so returns status, headers, data
           for the system server_error page."""
        # The tuple s_h_data is the tuple to return, start with it as None
        s_h_data = None
        try:
            if 'PATH_INFO' in environ:
                path = environ['PATH_INFO'].lower()
            else:
                raise ServerError(message="Invalid path", code=9034)

            # the path must start with this root project url
            if (path.find(self.url) != 0) and (path + "/" != self.url):
                # path does not start with the root, so send URL NOT FOUND
                return self._url_not_found(environ, path, lang)

            # This is the root project, check if the call is for a page in any sub project
            for proj, projurl in self._subproject_paths.items():
                if (path.find(projurl) == 0) or (path + "/" == projurl):
                    # this url is within a sub project
                    subproj = self.subprojects[proj]
                    s_h_data = subproj.proj_respond(environ, projurl, path, lang, received_cookies)
                    break
            else:
                # the call is for a page in this root project
                s_h_data = self.proj_respond(environ, self.url, path, lang, received_cookies)

            if s_h_data is None:
                # No page to return has been found, 
                return self._url_not_found(environ, path, lang)

            if s_h_data[2] is None:
                # No page data has been given
                return self._url_not_found(environ, path, lang)

        except ServerError as e:
            # if debug is enabled, expand the exception message to include the exception trace
            if skiboot.get_debug():
                e.message += "\n"
                exc_type, exc_value, exc_traceback = sys.exc_info()
                str_list = traceback.format_exception(exc_type, exc_value, exc_traceback)
                for item in str_list:
                    e.message += item
            # ServerError has occurred, return the server error page
            page = self._system_page("server_error")
            if (not page) or (page.page_type != "TemplatePage"):
                # return the default server error page
                return self.default_server_error_page(e.message, e.code)
            # import any sections
            page.import_sections()
            # show message passed by the exception
            page.show_error([e.errormessage])
            # if ServerError code, set it into the widget
            if e.code:
                if e.section:
                    page_data = {(e.section, e.widget, 'code'):str(e.code)}
                elif e.widget:
                    page_data = {(e.widget, 'code'):str(e.code)}
                elif page.default_error_widget.s:
                    page_data = {(page.default_error_widget.s, page.default_error_widget.w, 'code'):str(e.code)}
                elif page.default_error_widget.w:
                    page_data = {(page.default_error_widget.w, 'code'):str(e.code)}
                else:
                    page_data = None
                if page_data:
                    page.set_values(page_data)
            # update head and body parts
            page.update(environ, {}, lang, e.ident_list)
            status, headers = page.get_status()
            data = page.data()
            if not data:
                return self.default_server_error_page(e.message, e.code)
            # return page data
            s_h_data = e.status, headers, data

        return s_h_data



    def proj_respond(self, environ, projurl, path, lang, received_cookies):
        """Gets any received form data, and parses the ident field if present to find the caller page and ident_data
           Calls start call, and depending on the returned page, calls the project status_headers_data method"""

        caller_page = None
        ident_data = None

        rawformdata = cgi.FieldStorage(fp=environ['wsgi.input'], environ=environ)

        # if ident present in the rawformdata it should consist of project_pagenumber_identdata
        # and so the caller page can be found from the project_pagenumber
        # identdata may not exist

        try:
            # get the caller_page, from the ident field of submitted data
            # which could be None if no ident received

            if rawformdata and ('ident' in rawformdata):
                if not hasattr(rawformdata['ident'], 'value'):
                    raise ValidateError(message="Form data not accepted, caller page ident not recognised")
                else:
                    # rawformdata has 'ident' with attribute 'value'
                    # get the caller page ident, and the ident_data received from the 'ident' field
                    # which will be project_pagenumber_b64encodeddata

                    # Note: caller_page could belong to another project, so get it using ident.item() method
                    # which will query the right project

                    # project name, number and b64 encoded data should all be ascii characters passing this test
                    if _AN64.search(rawformdata['ident'].value):
                        raise ValidateError(message="Form data not accepted, caller page ident not recognised")

                    ident_parts = rawformdata['ident'].value.split('_', 2)
                    ident_items = len(ident_parts)
                    try:
                        if ident_items == 2:
                            caller_page = skiboot.Ident(ident_parts[0], int(ident_parts[1])).item()
                        elif ident_items == 3:
                            caller_page = skiboot.Ident(ident_parts[0], int(ident_parts[1])).item()
                            b64binarydata = ident_parts[2].encode('ascii') # get the submitted data and convert to binary
                            # add padding
                            b64binarydata = b64binarydata + b"=" * (4-len(b64binarydata)%4)
                            ident_data = urlsafe_b64decode(b64binarydata).decode('ascii') # b64 decode, and convert to string
                    except Exception:
                        caller_page = None
                    if caller_page is None:
                        raise ValidateError(message="Form data not accepted, (received ident is not valid)")
                    if caller_page.page_type != 'TemplatePage':
                        raise ValidateError(message="Form data not accepted, (caller page ident is not a template page)")

        except ValidateError as e:
            page = self._system_page("validate_error")
            if (not page) or (page.page_type != "TemplatePage"):
                return self.default_validate_error_page(e.message)
            # import any sections
            page.import_sections()
            # show message passed by the exception
            page.show_error([e.errormessage])
            # update head and body parts
            page.update(environ, {}, lang, e.ident_list)
            status, headers = page.get_status()
            data = page.data()
            if not data:
                return self.default_validate_error_page(e.message)
            # return page data
            return e.status, headers, data


        # so caller_page could be either given, or could be None

        # get the called ident, could be None
        ident = self.page_ident_from_path(projurl, path)

        if caller_page:
            caller_page_ident = caller_page.ident
        else:
            caller_page_ident = None

        # now call the proj_start_call function which creates a skicall object and
        # calls the users start_call function, which could return a different page ident, or None

        pident, skicall = self.proj_start_call(environ,
                                               path,
                                               ident,
                                               caller_page_ident,
                                               received_cookies,
                                               ident_data,
                                               lang)

        # parse the page ident returned

        if pident is None:
            # URL NOT FOUND and start_call does not divert
            return None

        # pident is the ident of the diverted page or a label or url string

        # get the page from pident
        if isinstance(pident, str):
            # either a label, or url
            if '/' in pident:
                # get redirector page
                return self._redirect_to_url(pident, environ, skicall.call_data, skicall.page_data, skicall.lang)
            else:
                # no '/' in pident so must be a label
                pident = skiboot.find_ident_or_url(pident, self._proj_ident)
                if not pident:
                    raise ServerError(message="Returned page ident from start_call not recognised", code=9036)
                if isinstance(pident, str):
                    # must be a url, get redirector page
                    return self._redirect_to_url(pident, environ, skicall.call_data, skicall.page_data, skicall.lang)

        # so pident must be an ident
        if not isinstance(pident, skiboot.Ident):
            raise ServerError(message="Invalid ident returned from start_call", code=9037)

        # pident is the ident returned from start_call, may be in a different project
        page = pident.item()
        if page is None:
            raise ServerError(message="Invalid ident returned from start_call", code=9038)
        if page.page_type == 'Folder':
            page = page.default_page
            if not page:
                raise ServerError(message="Invalid ident returned from start_call", code=9039)

        # read any submitted data from rawformdata, and place in form_data
        try:
            form_data = {}
            if rawformdata and (caller_page is not None) and (page.page_type == "RespondPage"):
                form_data = self.read_form_data(rawformdata, caller_page)

            # so form_data only available if
                # rawformdata has been submitted
                # and caller_page is known, so widgets can be extracted
                # and the destination is a RespondPage

                # otherwise form_data is empty (though rawformdata is retained)
        

            # dependent on wether the requested page is in this project or a sub project,
            # call status_headers_data() to find the final page to return to the client

            # ident_list retains list of idents called during a call to ensure no circulating calls
            ident_list = []
            # initially no errors, e_list is a list of errors to be shown
            e_list = []

            if page.ident.proj != self._proj_ident:
                # page returned from start_call is in another project
                subproj = self.subprojects.get(page.ident.proj)
                return subproj.status_headers_data(skicall, environ, received_cookies, rawformdata, caller_page, page, ident_list, e_list, form_data)
                
            # call status_headers_data to return status, headers and data to the top script
            return self.status_headers_data(skicall, environ, received_cookies, rawformdata, caller_page, page, ident_list, e_list, form_data)

        except ValidateError as e:
            page = self._system_page("validate_error")
            if (not page) or (page.page_type != "TemplatePage"):
                return self.default_validate_error_page(e.message)
            # import any sections
            page.import_sections()
            # show message passed by the exception
            page.show_error([e.errormessage])
            # update head and body parts
            page.update(environ, skicall.call_data, skicall.lang, e.ident_list)
            status, headers = page.get_status()
            data = page.data()
            if not data:
                return self.default_validate_error_page(e.message)
            # return page data
            return e.status, headers, page.data()



    def proj_start_call(self, environ, path, ident, caller_ident, received_cookies, ident_data, lang):
        """Creates a skicall object and calls the users start_call function
           ident is the ident of the page being called, could be None if not recognised
           Returns new called_ident, and the skicall object"""

        if not caller_ident:
            tuple_caller_ident = ()
        else:
            tuple_caller_ident = caller_ident.to_tuple()

        if ident is None:
            called_ident = None
        else:
            called_ident = ident.to_tuple()

        try:
            # create the SkiCall object
            skicall = SkiCall(environ = environ,
                              path = path,
                              project = self._proj_ident,
                              rootproject = self.rootproject,
                              caller_ident = tuple_caller_ident,
                              received_cookies = received_cookies,
                              ident_data = ident_data,
                              lang = lang,
                              proj_data = self.proj_data)

            # the skicall object is changed in place, with call_data and page_data
            # being set by the users own start_call function
            new_called_ident = self.start_call(called_ident, skicall)

            # convert returned tuple to an Ident object
            if isinstance(new_called_ident, int):
                new_called_ident = (self._proj_ident, new_called_ident)
            if isinstance(new_called_ident, tuple):
                new_called_ident = skiboot.make_ident(new_called_ident, self._proj_ident)
            # could be a label
        except ServerError as e:
            raise e
        except Exception as e:
            message = "Invalid exception in start_call function."
            raise ServerError(message, code=9040) from e
        return new_called_ident, skicall


    def read_form_data(self, rawformdata, caller_page):
        """Reads raw form data from the environ and returns a dictionary with keys as skiboot.WidgField objects and values as
           the form values.  Where input fields have indexed names, the skiboot.WidgField object still
           has i set to empty string, but the value is given as a dictionary with indexes as keys"""
        # rawformdata is the data obtained from environ
        # form_data is a dictionary of data returned, without the caller ident
        # and after a set of checks
        if not rawformdata:
            return {}
        if not caller_page:
            return {}
        form_data = {}
        for field in rawformdata.keys():
            # get fields and values from the rawformdata and store them in form_data
            # with keys as skiboot.WidgField objects, and values as field values
            # in the case of indexed fields, the values are dictionaries
            if field == 'ident':
                continue
            if ':' not in field:
                # All widgfields have a : in them to separate widget name from field name
                raise ValidateError(message="Form data not accepted, (invalid field %s)" % (field,))
            widgfield = skiboot.make_widgfield(field)
            # get fields and values from the rawformdata and store them in form_data
            widget = caller_page.copy_widget_from_name(widgfield.s, widgfield.w)
            if widget is None:
                raise ValidateError(message="Form data not accepted, (unexpected field %s)" % (field,))
            if isinstance(rawformdata[field], list):
                # fieldvalue is a list of items
                fieldvalue = [ item.value.strip() for item in rawformdata[field] ]
            else:
                fieldvalue = rawformdata[field].value.strip()
            if widget.is_senddict(widgfield.f):
                # field sends a dictionary, must have an index appended to the name
                # this part removes the index from the field name, and creates a form value of a dictionary with the index as keys
                fieldindex = widgfield.i
                if not fieldindex:
                    raise ValidateError(message="Form data not accepted, (invalid dictionary field %s)" % (field,))
                widgfieldnoindex = widgfield._replace(i='')
                if widgfieldnoindex in form_data:
                    form_data[widgfieldnoindex][fieldindex] = fieldvalue
                else:
                    form_data[widgfieldnoindex] = {fieldindex:fieldvalue}
            else:
                if widgfield.i:
                    raise ValidateError(message="Form data not accepted, (unexpected dictionary field %s)" % (field,))
                form_data[widgfield] = fieldvalue
        return form_data



    def status_headers_data(self, skicall, environ, received_cookies, rawformdata, caller_page, page, ident_list, e_list, form_data):
        """calls responders until it can return status, headers, page.data()"""

        try:
            while page.page_type == 'RespondPage':
                ident = page.ident
                if page.responder is None:
                    raise ServerError(message="Respond page %s does not have any responder set" % (page.url,), code=9042)
                try: 
                    page = page.call_responder(skicall, form_data, caller_page, ident_list, rawformdata)
                    if isinstance(page, str):
                        # must be a url
                        skicall.call_data.clear()
                        skicall.page_data.clear()
                        # get redirector page
                        return self._redirect_to_url(page, environ, skicall.call_data, skicall.page_data, skicall.lang)
                except PageError as ex:
                    # a jump to a page has occurred, with a list of errors
                    page = ex.page
                    if isinstance(page, str):
                        # must be a url
                        skicall.call_data.clear()
                        skicall.page_data.clear()
                        # get redirector page
                        return self._redirect_to_url(page, environ, skicall.call_data, skicall.page_data, skicall.lang)
                    if page.ident == ident:
                        raise ServerError(message="Invalid Failure page: can cause circulating call", code=9045)
                    if page.ident in ident_list:
                        raise ServerError(message="Invalid Failure page: can cause circulating call", code=9043)
                    # show the list of errors on the page
                    e_list = ex.e_list
                except GoTo as ex:
                    if ex.clear_submitted:
                        form_data.clear()
                    if ex.clear_page_data:
                        skicall.page_data.clear()
                    if ex.clear_errors:
                        ex.e_list = []
                    target = skiboot.find_ident_or_url(ex.target, ex.proj_ident)
                    # target is either an Ident, a URL or None
                    if not target:
                        raise ServerError(message="GoTo exception target not recognised", code=9044)
                    if isinstance(target, skiboot.Ident):
                        if target == ident:
                            raise ServerError(message="GoTo exception page ident %s invalid, can cause circulating call" % (target,), code=9045)
                        if target in ident_list:
                            raise ServerError(message="GoTo exception page ident %s invalid, can cause circulating call" % (target,), code=9046)
                        page = target.item()
                        if not page:
                            raise ServerError(message="GoTo exception page ident %s not recognised" % (target,), code=9047)
                        if page.page_type == 'Folder':
                            raise ServerError(message="GoTo exception page ident %s is a Folder, must be a page." % (target,), code=9048)
                    else:
                        # target is a URL
                        skicall.call_data.clear()
                        return self._redirect_to_url(target, environ, skicall.call_data, skicall.page_data, skicall.lang)
                    
                    # A divert to a fail page may lead to a GoTo exception which can therefore
                    # have an e_list
                    # show the list of errors on the page
                    e_list = ex.e_list

                # it is possible that a jump to a page in another project has been made
                if page.ident.proj != self._proj_ident:
                    subproj = skiboot.getproject(proj_ident=page.ident.proj)
                    return subproj.status_headers_data(skicall, environ, received_cookies, rawformdata, caller_page, page, ident_list, e_list, form_data)
                
        except (ServerError, ValidateError) as e:
            e.ident_list = ident_list
            raise e

        # the page to be returned to the client is now 'page'
        # and 'e_list' is a list of errors to be shown on it

        # call the user function end_call
        try:
            skicall.project = self._proj_ident
            skicall.proj_data = self.proj_data
            skicall.rootproject = self.rootproject
            try:
                session_string = self.end_call(page.ident.to_tuple(), page.page_type, skicall)
                if session_string:
                    # set cookie in target_page
                    page.session_cookie = "Set-Cookie", "%s=%s; Path=%s" % (skicall.project, session_string, skiboot.root_project().url)
            except FailPage as e:
                page.show_error([e.errormessage])
            finally:
                if skicall._lang_cookie:
                    page.language_cookie = skicall._lang_cookie
        except GoTo as e:
            raise ServerError("Invalid GoTo exception in end_call", code=9049) from e
        except Exception as e:
            message = "Invalid exception in end_call function."
            raise ServerError(message, code=9050) from e

        # import any sections
        page.import_sections(skicall.page_data)
        if e_list:
            # show the list of errors on the page
            page.show_error(e_list)
        # now set the widget fields
        if skicall.page_data:
            page.set_values(skicall.page_data)
        page.update(environ, skicall.call_data, skicall.lang, ident_list)
        status, headers = page.get_status()
        return status, headers, page.data()



    def clear_cache(self):
        "clear the cache of paths"
        self._paths = {}


    def load_from_json(self):
        "Loads project with data saved in project.json file"
        projectdict = read_json.create_project(self._proj_ident, self.projectfiles)
        self.default_language = projectdict['default_language']
        self.brief = projectdict['brief']
        self.version = projectdict['version']
        self.special_pages = projectdict['specialpages']
        self.sections = projectdict['sections']
        self.root = projectdict['siteroot']
        itemlist = projectdict['itemlist']
        self.identitems = {}
        if itemlist:
            for item in itemlist:
                self.identitems[item.ident] = item


    @property
    def max_ident_num(self):
        "Returns the maximum identnumber currently in use"
        maxnum = 0
        for ident in self.identitems:
            if ident.num > maxnum:
                    maxnum = ident.num
        return maxnum

    def list_section_names(self):
        "Returns a list of section names, alphabetacily ordered"
        if not self.sections:
            return []
        s = [ name for name in self.sections ]
        s.sort()
        return s

    def section(self, section_name, makecopy=True):
        "Returns a section, or a deep copy of a section, or None if the section name is not found"
        if section_name not in self.sections:
            return
        section = self.sections[section_name]
        if not makecopy:
            return section
        if section is None:
            return None
        return copy.deepcopy(section)
 

    def add_section(self, name, section):
        "Adds a section to the project, returns section.change uuid"
        # and save the section
        section.widgets = {}
        section.section_places = {}  # currently unused
        embedded = (name, '', None)
        section.set_idents(name, section.widgets, section.section_places, embedded)
        # set validators in section
        section.load_validator_scriptlinks()
        # set the section change number
        section.change = uuid.uuid4().hex
        self.sections[name] = section
        return section.change


    def delete_section(self, name):
        "Deletes a section"
        if name in self.sections:
            del self.sections[name]

    @property
    def ident_numbers(self):
        "return a list of ident numbers"
        num_list = [ ident.num for ident in self.identitems ]
        # insert the root
        num_list.insert(0,0)
        num_list.sort()
        return num_list

    def __getitem__(self, ident):
        """given an Ident, or a string version of ident, return page or folder. 
              If folder or respond page return the item, any other page, return a deep copy
              of the item. If item not found, return None"""
        ident = skiboot.Ident.to_ident(ident, proj_ident=self._proj_ident)
        if ident is None:
            return
        if ident.proj != self._proj_ident:
            return
        if ident.num == 0:
            return self.root
        if ident not in self.identitems:
            return
        item = self.identitems[ident]
        if item is None:
            return
        if item.page_type == 'Folder':
            return item
        if item.page_type == 'RespondPage':
            return item
        return copy.deepcopy(item)

    def add_item(self, parent_ident, item, ident=None):
        """Adds a new page or folder to the project, returns the item ident"""

        # check ident
        if ident is None:
            ident = skiboot.Ident(self._proj_ident, self.max_ident_num+1)
        else:
            ident = skiboot.make_ident(ident, self._proj_ident)
            if ident is None:
                raise ServerError(message="Sorry. Invalid ident", code=9000)
            if ident.num == 0:
                # cannot add the root folder
                raise ServerError(message="Sorry. Unable to add a new root", code=9001)
            if ident.proj != self._proj_ident:
                # must be in this project
                raise ServerError(message="Sorry. Invalid ident", code=9002)
            if ident in self.identitems:
                # ident must not exist
                raise ServerError(message="Sorry. The given ident already exists", code=9003)

        # check parent folder
        if parent_ident.proj != self._proj_ident:
            raise ServerError(message="Invalid parent ident", code=9004)
        parent = self.get_item(parent_ident)
        if parent is None:
            raise ServerError(message="Parent folder not found: Error in add_item method of Project class.", code=9005)
        if parent.page_type != 'Folder':
            raise ServerError(message="Parent not a folder", code=9006)

        if item.name in parent.pages:
            raise ServerError(message="Sorry, a page with that name already exists in the parent folder", code=9007)
        if item.name in parent.folders:
            raise ServerError(message="Sorry, a folder with that name already exists in the parent folder", code=9008)

        # set the item ident
        item.ident = ident

        # set this item name into the parent
        if item.page_type == 'Folder':
            if parent.restricted:
                item.set_restricted()
            parent.folders[item.name] = item.ident
        # if the page is a template or svg page, then set its idents
        # and store its widgets in the page's widgets dictionary and sectionplaceholders
        # in the page's sections directory
        elif (item.page_type == 'TemplatePage') or (item.page_type == 'SVG'):
            item.set_idents()
            parent.pages[item.name] = item.ident
            # now set validator modules in page
            if item.page_type == 'TemplatePage':
                item.load_validator_scriptlinks()
        else:
            parent.pages[item.name] = item.ident

        # set the parent change value
        parent.change = uuid.uuid4().hex
        item.parentfolder = parent

        # and finally, add the item
        self.identitems[item.ident] = item
        self.clear_cache()
        return item.ident


    def delete_item(self, itemident):
        """Deletes the page or folder with the given ident from the database."""
        if itemident.num == 0:
            # cannot delete the root folder
            raise ServerError(message="Cannot delete the root folder", code=9009)
        if itemident.proj != self._proj_ident:
            # Must belong to this project
            raise ServerError(message="Cannot delete this item (does not belong to this project)", code=9010)
        if itemident not in self.identitems:
            raise ServerError(message="Item not found", code=9011)
        # get the item
        item = self.identitems[itemident]
        # get the items parent folder
        parentfolder = item.parentfolder
        if item.name in parentfolder.pages:
            del parentfolder.pages[item.name]
        if item.name in parentfolder.folders:
            del parentfolder.folders[item.name]
        parentfolder.change = uuid.uuid4().hex
        # del the item
        del self.identitems[itemident]
        self.clear_cache()

    def delete_folder_recursively(self, itemident):
        """Deletes the folder and contents with the given ident from the database.
           returns parentfolder number and change when done, raises ServerError on failure"""
        if itemident.num == 0:
            # cannot delete the root folder
            raise ServerError(message="Cannot delete the root folder", code=9009)
        if itemident.proj != self._proj_ident:
            # Must belong to this project
            raise ServerError(message="Cannot delete this item (does not belong to this project)", code=9010)
        if itemident not in self.identitems:
            raise ServerError(message="Item not found", code=9011)
        # get the item
        item = self.identitems[itemident]
        # get the items parent folder
        parentfolder = item.parentfolder
        if item.name in parentfolder.pages:
            raise ServerError(message="The item is not a Folder")
        if item.name not in parentfolder.folders:
            raise ServerError(message="The item to delete has not been found")
        # recursive delete this item and all contents
        self._do_delete_folder_recursively(itemident)
        # create change uuid for parentfolder
        change = uuid.uuid4().hex
        num = parentfolder.ident.num
        parentfolder.change = change
        self.clear_cache()
        return num, change


    def _do_delete_folder_recursively(self, folderident):
        "Used by delete_folder_recursively to rcursively delete"
        # get the folder
        folder = self.identitems[folderident]
        page_idents = folder.page_idents()
        folder_idents = folder.folder_idents()

        # get the parent folder
        parentfolder = folder.parentfolder

        # delete all subfolders
        for folder_ident in folder_idents:
            self._do_delete_folder_recursively(folder_ident)

        # delete all pages in the folder
        for page_ident in page_idents:
            page = self.identitems[page_ident]
            del folder.pages[page.name]
            del self.identitems[page_ident]

        # and finally delete itself
        del parentfolder.folders[folder.name]
        # del this item from identitems
        del self.identitems[folderident]


    def save_item(self, item, new_parent_ident=None):
        """Saves the page or folder - used to save an altered item, not to add a new one
           If new_parent_ident is not None, indicates the item has moved to a different folder
           returns the items new change uuid"""
        if item.page_type == 'Folder':
            return self.save_folder(item, new_parent_ident)
        else:
            return self.save_page(item, new_parent_ident)


    def save_page(self, item, new_parent_ident=None):
        """Saves the page - used to save an altered page, not to add a new one
           If new_parent_ident is not None, indicates the page has moved to a different folder
           Returns the new page.change uuid"""
        if item.page_type == 'Folder':
            raise ServerError(message="Invalid item, not a page.", code=9012)
        item_ident = item.ident
        if item_ident is None:
            raise ServerError(message="Unable to save page - no ident set", code=9013)
        if self._proj_ident != item_ident.proj:
            raise ServerError(message="Unable to save page - invalid ident", code=9014)
        if item_ident not in self.identitems:
            raise ServerError(message="This page ident does not exist", code=9015)
        old_parent = self.identitems[item_ident].parentfolder
        old_name = self.identitems[item_ident].name
        if new_parent_ident is not None:
            # So its a parent folder change
            if new_parent_ident.num == 0:
                # new parent is root
                new_parent = self.root
            else:
                new_parent = self.identitems[new_parent_ident]
            if new_parent == old_parent:
                new_parent_ident = None
        if (item.page_type == 'TemplatePage') or (item.page_type == 'SVG'):
            item.set_idents()
        # now set validator modules in page
        if item.page_type == 'TemplatePage':
            item.load_validator_scriptlinks()
        item.change = uuid.uuid4().hex
        if (old_name == item.name) and (new_parent_ident is None):
            # no folder change
            self.identitems[item_ident] = item
            self.clear_cache()
            return item.change
        if new_parent_ident is None:
            # so just a name change
            if item.name in old_parent.pages:
                raise ServerError(message="Sorry, a page with that name already exists", code=9016)
            if item.name in old_parent.folders:
                raise ServerError(message="Sorry, a folder with that name already exists", code=9017)
            if old_name in old_parent.pages:
                del old_parent.pages[old_name]
            old_parent.pages[item.name] = item_ident
            old_parent.change = uuid.uuid4().hex
            self.identitems[item_ident] = item
            self.clear_cache()
            return item.change
        # change of folder
        if item.name in new_parent.pages:
            raise ServerError(message="Sorry, a page with that name already exists", code=9018)
        if item.name in new_parent.folders:
            raise ServerError(message="Sorry, a folder with that name already exists", code=9019)
        if old_name in old_parent.pages:
            del old_parent.pages[old_name]
            old_parent.change = uuid.uuid4().hex
        new_parent.pages[item.name] = item_ident
        new_parent.change = uuid.uuid4().hex
        item.parentfolder = new_parent
        self.identitems[item_ident] = item
        self.clear_cache()
        return item.change


    def save_folder(self, item, new_parent_ident=None):
        """Saves the folder - used to save an altered folder, not to add a new one
           If new_parent_ident is not None, indicates the folder has moved to a different parent folder
           Returns the new folder.change uuid"""
        if item.page_type != 'Folder':
            raise ServerError(message="Invalid item, not a folder.", code=9020)
        item_ident = item.ident
        if item_ident is None:
            raise ServerError(message="Unable to save folder - no ident set", code=9021)
        if self._proj_ident != item_ident.proj:
            raise ServerError(message="Unable to save folder - invalid ident", code=9022)
        if item_ident.num == 0:
            if new_parent_ident:
                raise ServerError(message="Root folder cannot have new parent", code=9023)
            item.change = uuid.uuid4().hex
            self.root = item
            self.clear_cache()
            return item.change
        if item_ident not in self.identitems:
            raise ServerError(message="This folder ident does not exist", code=9024)
        old_parent = self.identitems[item_ident].parentfolder
        old_name = self.identitems[item_ident].name
        if new_parent_ident is not None:
            # So its a parent folder change
            if new_parent_ident.num == 0:
                # new parent is root
                new_parent = self.root
            else:
                new_parent = self.identitems[new_parent_ident]
            if new_parent == old_parent:
                new_parent_ident = None
        item.change = uuid.uuid4().hex
        if (old_name == item.name) and (new_parent_ident is None):
            # no parent folder change
            self.identitems[item_ident] = item
            self.clear_cache()
            return item.change
        if new_parent_ident is None:
            # so just a name change
            if item.name in old_parent.pages:
                raise ServerError(message="Sorry, a page with that name already exists", code=9025)
            if item.name in old_parent.folders:
                raise ServerError(message="Sorry, a folder with that name already exists", code=9026)
            if old_name in old_parent.folders:
                del old_parent.folders[old_name]
            old_parent.folders[item.name] = item_ident
            old_parent.change = uuid.uuid4().hex
            self.identitems[item_ident] = item
            self.clear_cache()
            return item.change
        # change of folder
        # A folder cannot be moved into a sub folder of itself

        folder_list = new_parent.parent_list()
        # folder list is a list of (name, identnumber) starting at root
        folder_ident_numbers = [ identnumber for name,identnumber in folder_list ]
        if item.ident.num in folder_ident_numbers:
            # item is a parent of new_parent
            raise ServerError(message="Sorry, a folder cannot be moved into a subfolder of itself.", code=9027)
        if item.name in new_parent.pages:
            raise ServerError(message="Sorry, a page with that name already exists", code=9028)
        if item.name in new_parent.folders:
            raise ServerError(message="Sorry, a folder with that name already exists", code=9029)
        if old_name in old_parent.folders:
            del old_parent.folders[old_name]
            old_parent.change = uuid.uuid4().hex
        new_parent.folders[item.name] = item_ident
        new_parent.change = uuid.uuid4().hex
        item.parentfolder = new_parent
        self.identitems[item_ident] = item
        self.clear_cache()
        return item.change



    def get_item(self, ident):
        """given an ident (Ident object or integer), return a
           folder or page from the database, if not found, return None.
           Note: the item is returned without copying and without sections imported"""
        if isinstance(ident, int):
            ident = skiboot.Ident(self._proj_ident, ident)
        elif ident.proj != self._proj_ident:
            return None
        if ident.num == 0:
            return self.root
        if ident not in self.identitems:
            return None
        return self.identitems[ident]


    def __iter__(self):
        "This iterator does not return the root folder"
        for ident in self.identitems:
            yield ident

    def __contains__(self, item):
        "Checks if this project contains folder, page or ident"
        if hasattr(item, 'ident'):
            ident = item.ident
        else:
            ident = skiboot.Ident.to_ident(item, self._proj_ident)
        if ident in self.identitems:
            return True
        if (ident.proj == self._proj_ident) and (ident.num == 0):
            return True
        return False

    def __len__(self):
        "This length does not include the root"
        return len(self.identitems)

    def __bool__(self):
        return True

    def set_special_page(self, label, target):
        "Sets a special page"
        if not label:
            raise ServerError(message="Sorry, a special page label must be given", code=9030)
        if not target:
            raise ServerError(message="Sorry, a label target must be given", code=9031)
        if isinstance(target, str) and ( '/' in target ):
                # item is a url
                item = target
        elif isinstance(target, str) and ( ',' in target ) and not (target.split(',')[1].isdigit()):
            if len(target.split(',')) == 2:
                # item points to a subproject label
                item = target
        else:
            item = skiboot.make_ident(target, self._proj_ident)
            if not item:
                raise ServerError(message="Sorry, the page target is not recognised", code=9032)
        self.special_pages[label] = item


    def delete_special_page(self, label):
        "Deletes a special page"
        if label in self.special_pages:
            del self.special_pages[label]

    def _system_page(self, label):
        """Returns the system page with the given label, if not found, returns None"""
        # label must be one of the system pages
        if label not in skiboot.sys_list():
            return
        if label not in self.special_pages:
            return
        target = self.special_pages[label]
        ident = skiboot.find_ident(target, proj_ident=self._proj_ident)
        if ident is None:
            return
        return ident.item()


    def labels(self):
        "Returns the special pages dictionary as a dictionary of labels with url's and tuples"
        labels_dict = {}
        for key, val in self.special_pages.items():
            if isinstance(val, str):
                labels_dict[key] = val
            else:
                labels_dict[key] = val.to_tuple()
        return labels_dict


    def label_value(self, key):
        "Returns the url or tuple associated with the label"
        val = self.special_pages.get(key)
        if val is None:
            return
        labels_dict = {}
        if isinstance(val, str):
            return val
        else:
            return val.to_tuple()


    def _redirect_to_url(self, url, environ, call_data, page_data, lang):
        "Return status, headers, page.data() of the redirector page, with fields set to url"
        if '/' not in url:
            raise ServerError(message="Invalid target url, must contain at least one /", code=9033)
        page = self._system_page('redirector')
        if (not page) or (page.page_type != "TemplatePage"):
            page_text = "<!DOCTYPE HTML>\n<html>\n<p>Page Redirect request. Please try: <a href=%s>%s</a></p>\n</html>" % (url, url)
            return '200 OK', [('content-type', 'text/html')], [page_text.encode('ascii', 'xmlcharrefreplace')]
        # create an ErrorMessage with the url as the message
        err = ErrorMessage(message=html.escape(url))
        # import any sections
        page.import_sections()
        page.show_error(error_messages=[err])
        # update head and body parts
        if page_data:
            page.set_values(page_data)
        page.update(environ, call_data, lang)
        status, headers = page.get_status()
        return status, headers, page.data()


    def _url_not_found(self, environ, path, lang):
        "Used to return the url not found page"
        page = self._system_page("url_not_found")
        page_content = "<!DOCTYPE HTML>\n<html>\nERROR:UNKNOWN URL\n</html>".encode('ascii', 'xmlcharrefreplace')
        if not page:
            return '404 Not Found', [('content-type', 'text/html')], [page_content]
        if page.page_type == "TemplatePage":
            # create an ErrorMessage with the path as the message
            err = ErrorMessage(message=html.escape(path))
            # import any sections
            page.import_sections()
            page.show_error(error_messages=[err])
        if (page.page_type == "TemplatePage") or (page.page_type == "FilePage"):
            # update head and body parts
            page.update(environ, {}, lang)
            status, headers = page.get_status()
            page_data = page.data()
            if page_data:
                return '404 Not Found', headers, page_data
        # if any other type of page, or no page content, then return this
        return '404 Not Found', [('content-type', 'text/html')], [page_content]


    def default_server_error_page(self, message='', code=0):
        "Given a ServerError exception, return a default status,headers,data"
        text_start = "<!DOCTYPE HTML>\n<html>\n<p>SERVER ERROR</p>\n<p>Error code : %s</p>\n" % (code,)
        if message:
            page_text = text_start + "<pre>%s</pre>\n</html>" % (html.escape(message),)
        else:
            page_text = text_start + "</html>"
        return '500 Internal Server Error', [('content-type', 'text/html')], [page_text.encode('ascii', 'xmlcharrefreplace')]


    def default_validate_error_page(self, message):
        "Given a ValidateError exception, return a default status,headers,data"
        if message:
            page_text = "<!DOCTYPE HTML>\n<html>\n<p>VALIDATION ERROR</p>\n<p>%s</p>\n</html>" % (html.escape(message),)
        else:
            page_text = "<!DOCTYPE HTML>\n<html>\n<p>VALIDATION ERROR</p>\n</html>"
        return '400 Bad Request', [('content-type', 'text/html')], [page_text.encode('ascii', 'xmlcharrefreplace')]


    def page_ident_from_path(self, projurl, path):
        """Tests if ident exists in the cache, return it, if not, call self.root.page_ident_from_path
           and cach the result, then return the ident. If no ident found, or if ident within a restricted folder, return None."""
        if path in self._paths:
            return self._paths[path]
        ident = None
        strip_path = path.strip("/")
        if not strip_path:
            pathlist = []
        else:
            pathlist = strip_path.split("/")
        strip_projurl = projurl.strip("/")
        # The projurl must be removed from the pathlist before the call to self.root.page_ident_from_path()
        if (not strip_projurl):
            # no projurl to remove
            ident = self.root.page_ident_from_path(self.identitems, pathlist)
        else:
            # strip_projurl may be something like "lib", remove the projurl from the pathlist
            projurl_list = strip_projurl.split("/")
            for item in projurl_list:
                if not pathlist:
                    # invalid call, the pathlist must start with the projurl
                    return
                if item == pathlist[0]:
                    pathlist.pop(0)
                else:
                    # invalid call, the pathlist must start with the projurl
                    return
            ident = self.root.page_ident_from_path(self.identitems, pathlist)
        if ident is not None:
            self._paths[path] = ident
        return ident


    @property
    def proj_ident(self):
        return self._proj_ident

    @property
    def subproject_paths(self):
        "property getter to return an ordered dictionary of sub project {ident:path,...}"
        if not self.rootproject:
            # sub project do not themselves contain further sub projects
            return collections.OrderedDict()
        return self._subproject_paths.copy()


    def list_of_subproject_idents(self):
        "Returns a list of subproject idents"
        return [i for i in self.subproject_paths]

    def add_project(self, proj, url=None):
        """Add a project to self, returns the url
           proj is the sub project application.
           This adds a reference to the project to the subproject_paths, returns the sub project path"""
        if not self.rootproject:
           raise ValidateError(message="Cannot add to a sub project")
        proj_id = proj.proj_ident

        # get a copy of the {proj_id:url} subproject_paths dictionary, and this projects url
        sub_paths = self._subproject_paths.copy()

        if (proj_id == self._proj_ident) or (proj_id in sub_paths):
            raise ValidateError(message="Project already exits")

        this_url = self.url

        if url is None:
            url = proj.url.strip("/").lower()
            if not url:
                url = proj_id.lower()
        else:
            url=url.strip("/").lower()
            if not url:
                raise ValidateError(message="Invalid URL passed to add_project, it must be a path longer than the root application path")
        url = "/" + url + "/"
        # Ensure url starts with this project url
        if not url.startswith(this_url):
            raise ValidateError(message="Invalid URL passed to add_project, it must be a path longer than the root application path")
        # add this ident and url to subproject_paths
        sub_paths[proj_id] = url
        # save new subproject_paths dictionary
        self._subproject_paths = collections.OrderedDict(sorted(sub_paths.items(), key=lambda t: len(t[1].strip("/").split("/")), reverse=True))
        # add the subproject to this project
        proj.rootproject = False
        proj.url = url
        proj._subproject_paths = collections.OrderedDict()
        proj.subprojects = {}
        self.subprojects[proj_id] = proj
        return url

    @property
    def root_ident(self):
        'provides a root_ident attribute'
        return skiboot.Ident(self._proj_ident, 0)

    def next_ident(self):
        "Returns next Ident available by incrementing the maximum existing ident number"
        return skiboot.Ident(self._proj_ident, self.max_ident_num+1)


class SkiCall(object):
    """SkiCall is the class of the skicall object which is created for each incoming
       call and is passed as an argument to the user functions"""

    def __init__(self, environ, path, project, rootproject, caller_ident, received_cookies, ident_data, lang, proj_data):

        self.environ = environ
        self.path = path
        self.project = project
        self.rootproject = rootproject
        self.caller_ident = caller_ident
        self.received_cookies = received_cookies
        self.ident_data = ident_data
        self._lang = lang
        self._lang_cookie = None
        self.proj_data = proj_data

        self._projectfiles = skiboot.projectfiles(project)

        self.ident_list = []
        self.submit_list = []
        self.submit_dict = {'error_dict':{}}
        self.call_data = {}
        self.page_data = {}


    @property
    def projectfiles(self):
        "Returns the projectfiles string"
        return self._projectfiles

    @property
    def lang(self):
        "Returns the lang tuple"
        return self._lang

    def get_language(self):
        "Returns the language string"
        return self._lang[0]

    def set_language(self, language):
        "Sets the language string and creates a language cookie with a persistance of 30 days"
        if language:
            self._lang = (language, self._lang[1])
            self._lang_cookie = "Set-Cookie", "language=%s; Path=%s; Max-Age=2592000" % (language_string, skiboot.root_project().url)

    language = property(get_language, set_language)

    @property
    def accesstextblocks(self):
        "Returns the project instance of the AccessTextBlocks class"
        this_project = skiboot.getproject(proj_ident=self.project)
        return this_project.textblocks

    def textblock(self, textref, project=None):
        """This method returns the textblock text, given a textblock reference string,
           If project is not given assumes this project, if given, project must exist as either the root,
           or a sub project of the root.
           If no textblock is found, returns None."""
        if project is None:
            project = self.project
        proj = skiboot.getproject(project)
        if proj is None:
            return
        return proj.textblocks.get_text(textref, self.lang)

    def label_value(self, label, project=None):
        """Given a label, returns the associated ident or URL
           If project is not given assumes this project, if given, project must exist as either the root,
           or a sub project of the root.
           If no label is found, returns None."""
        if project is None:
            project = self.project
        proj = skiboot.getproject(project)
        if proj is None:
            return
        return proj.label_value(label)


    def projectpaths(self):
        """Returns a dictionary of project name : project path

           This method returns a dictionary of project names as keys with the project paths as values."""
        all_projects = skiboot.project_register()
        return {proj_ident:proj.url for proj_ident, proj in all_projects.items()}








from collections import OrderedDict

from . import skiboot, excepts, tag


def page_to_OD(proj_ident, page):
    """Returns an OrderedDictionary of the page"""
    proj = skiboot.getproject(proj_ident)
    ident = page.ident
    page_dict = OrderedDict()
    page_dict["name"] = page.name
    page_dict["version"] = proj.version
    # stores the version of this skipole
    page_dict["skipole"] = skiboot.version()
    page_dict["ident"] = ident.num
    if page.brief:
        page_dict["brief"] = page.brief
    if page.page_type == 'FilePage':
        return _create_filepage(page_dict, page, ident, proj_ident)
    elif page.page_type == 'RespondPage':
        return _create_respondpage(page_dict, page, ident, proj_ident)
    elif page.page_type == 'CSS':
        return _create_csspage(page_dict, page, ident, proj_ident)
    elif page.page_type == 'JSON':
        return _create_jsonpage(page_dict, page, ident, proj_ident)
    elif page.page_type == 'SVG':
        return _create_svgpage(page_dict, page, ident, proj_ident)
    elif page.page_type == 'TemplatePage':
        return _create_templatepage(page_dict, page, ident, proj_ident)
    else:
        return OrderedDict()


def folder_to_OD(proj_ident, folder):
    """Returns an OrderedDictionary of the folder"""
    proj = skiboot.getproject(proj_ident)
    if isinstance(folder, skiboot.Ident):
        ident = folder
    else:
        ident = folder.ident
    if ident.num:
        folder_dict = _create_folder(ident, proj_ident)
        folder_name = folder.name
    else:
        folder_dict = _create_root_folder(proj_ident)
        folder_dict["ident"] = 0
        folder_dict.move_to_end("ident", last=False)
        folder_name = proj_ident + "_root"
    # stores the version of this skipole
    folder_dict["skipole"] = skiboot.version()
    folder_dict.move_to_end("skipole", last=False)
    folder_dict["version"] = proj.version
    folder_dict.move_to_end("version", last=False)
    folder_dict["name"] = folder_name
    folder_dict.move_to_end("name", last=False)
    return folder_dict


def project_to_OD(proj_ident=None):
    """Returns an OrderedDictionary of the project"""
    if proj_ident is None:
        proj_ident = skiboot.project_ident()
    # proj is the actual Project instance to be dumped
    proj = skiboot.getproject(proj_ident)

    # create a big ordered dictionary
    project = OrderedDict()

    project["url"] = proj.url
    project["default_language"] = proj.default_language
    brief = proj.brief
    if not brief:
        project["brief"] = "Project : %s" % (proj_ident,)
    else:
        project["brief"] = brief
    project["version"] = proj.version

    # stores the version of this skipole
    project["skipole"] = skiboot.version()

    specialpages = {}
    specpages = proj.special_pages
    if specpages:
        for label, target in specpages.items():
            if isinstance(target, skiboot.Ident):
                if target.proj == proj_ident:
                    specialpages[label] = target.num
                else:
                    specialpages[label] = str(target)
            else:
                specialpages[label] = target
        project["specialpages"] = OrderedDict(sorted(specialpages.items(), key=lambda t: t[0]))
    else:
        project["specialpages"] = OrderedDict()

    sections = OrderedDict()
    projsections = proj.sections
    if projsections:
        for section_name, section in projsections.items():
            sections[section_name] = section.outline(proj_ident)
        project["sections"] = OrderedDict(sorted(sections.items(), key=lambda t: t[0]))
    else:
        project["sections"] = OrderedDict()

    # add rootfolder to the project
    project["RootFolder"] = _create_root_folder(proj_ident)

    return project


def _create_root_folder(proj_ident):
    "Creates a dictionary for the root folder"
    if proj_ident is None:
        proj_ident = skiboot.project_ident()
    proj = skiboot.getproject(proj_ident)
    # rootfolder is the OrderedDict to return
    rootfolder = OrderedDict()
    root = proj.root
    brief = root.brief
    if not brief:
        rootfolder["brief"] = "Root folder"
    else:
        rootfolder["brief"] = brief
    default_page_name = root.default_page_name
    if default_page_name:
        rootfolder["default_page_name"] = default_page_name
    rootfolder["restricted"] = False
    # folders in root
    root_folders = OrderedDict()
    folders = root.folders
    if folders:
        for foldername, ident in folders.items():
            root_folders[foldername] = _create_folder(ident, proj_ident)
        rootfolder["folders"] = OrderedDict(sorted(root_folders.items(), key=lambda t: t[0]))
    else:
        rootfolder["folders"] = OrderedDict()
    # pages in root
    root_pages = {}
    pages = root.pages
    if pages:
        for pagename, ident in pages.items():
            root_pages[pagename] = _create_page(ident, proj_ident)
        rootfolder["pages"] = OrderedDict(sorted(root_pages.items(), key=lambda t: t[0]))
    else:
        rootfolder["pages"] = OrderedDict()
    return rootfolder


def _create_folder(ident, proj_ident):
    "Creates a dictionary for the folder"
    proj = skiboot.getproject(proj_ident)
    folder_dict = OrderedDict()
    folder = proj.identitems[ident]

    folder_dict["ident"] = ident.num
    if folder.brief:
        folder_dict["brief"] = folder.brief
    if folder.default_page_name:
        folder_dict["default_page_name"] = folder.default_page_name
    folder_dict["restricted"] = folder.restricted

    # folders in this folder
    sub_folders = {}
    folders = folder.folders
    if folders:
        for foldername, folderident in folders.items():
            sub_folders[foldername] = _create_folder(folderident, proj_ident)
        folder_dict["folders"] = OrderedDict(sorted(sub_folders.items(), key=lambda t: t[0]))
    else:
        folder_dict["folders"] = OrderedDict()

    # pages in this folder
    pages_dict = {}
    pages = folder.pages
    if pages:
        for pagename, pageident in pages.items():
            pages_dict[pagename] = _create_page(pageident, proj_ident)
        folder_dict["pages"] = OrderedDict(sorted(pages_dict.items(), key=lambda t: t[0]))
    else:
        folder_dict["pages"] = OrderedDict()

    return folder_dict


def _create_page(ident, proj_ident):
    "Creates a dictionary for the page"
    proj = skiboot.getproject(proj_ident)
    page_dict = OrderedDict()
    page = proj.identitems[ident]
    page_dict["ident"] = ident.num
    if page.brief:
        page_dict["brief"] = page.brief
    if page.page_type == 'FilePage':
        return _create_filepage(page_dict, page, ident, proj_ident)
    elif page.page_type == 'RespondPage':
        return _create_respondpage(page_dict, page, ident, proj_ident)
    elif page.page_type == 'CSS':
        return _create_csspage(page_dict, page, ident, proj_ident)
    elif page.page_type == 'JSON':
        return _create_jsonpage(page_dict, page, ident, proj_ident)
    elif page.page_type == 'SVG':
        return _create_svgpage(page_dict, page, ident, proj_ident)
    elif page.page_type == 'TemplatePage':
        return _create_templatepage(page_dict, page, ident, proj_ident)
    else:
        return OrderedDict()


def _make_dictionary(input_dict, proj_ident):
    """Used by other functions, takes input_dict and returns
       an ordered dict, with TextBlocks, idents converted for json compatability"""
    output_dict = OrderedDict()
    if not input_dict: return output_dict
    for key, val in input_dict.items():
        if val is None:
            output_dict[key] = None
        elif val == '':
            output_dict[key] = ""
        elif isinstance(val, list):
            output_dict[key] = _make_list(val, proj_ident)
        elif isinstance(val, dict):
            output_dict[key] = _make_dictionary(val, proj_ident)
        elif val is True:
            output_dict[key] = True
        elif val is False:
            output_dict[key] = False
        elif isinstance(val, tag.TextBlock):
            output_dict[key] = val.textref
        elif isinstance(val, skiboot.Ident):
            if val.proj == proj_ident:
                # ident is this project, put the number only
                output_dict[key] = val.num
            else:
                # ident is another project, put the full ident
                output_dict[key] = [val.proj, val.num]
        else:
            output_dict[key] = str(val)
    return output_dict


def _make_list(input_list, proj_ident):
    """Used by other functions, takes input_list and returns a list with items converted"""
    if not input_list: return []
    output_list = []
    for item in input_list:
        if item is None:
            output_list.append(None)
        elif item == '':
            output_list.append('')
        elif isinstance(item, list):
            output_list.append(_make_list(item, proj_ident))
        elif isinstance(item, dict):
            output_list.append(_make_dictionary(item, proj_ident))
        elif item is True:
            output_list.append(True)
        elif item is False:
            output_list.append(False)
        elif isinstance(item, skiboot.Ident):
            if item.proj == proj_ident:
                output_list.append(item.num)
            else:
                # ident is another project, put the full ident
                output_list.append([item.proj, item.num])
        else:
            output_list.append(str(item))
    return output_list


def _create_filepage(page_dict, page, ident, proj_ident):
    "Creates a dictionary for the filepage"
    page_args = OrderedDict()
    page_args["filepath"] = page.filepath
    page_args["enable_cache"] = page.enable_cache
    if page.mimetype:
        page_args["mimetype"] = page.mimetype

    page_dict["FilePage"] = page_args
    return page_dict


def _create_respondpage(page_dict, page, ident, proj_ident):
    "Creates a dictionary for the respondpage"
    page_args = OrderedDict()
    page_args["class"] = page.responder.__class__.__name__
    original_args = OrderedDict(sorted(page.responder.original_args().items(), key=lambda t: t[0]))
    page_args["original_args"] = _make_dictionary(original_args, proj_ident)
    original_fields = OrderedDict(sorted(page.responder.original_fields().items(), key=lambda t: t[0]))
    page_args["original_fields"] = _make_dictionary(original_fields, proj_ident)
    page_dict["RespondPage"] = page_args
    return page_dict


def _create_csspage(page_dict, page, ident, proj_ident):
    "Creates a dictionary for the page"
    page_args = OrderedDict()
    page_args["enable_cache"] = page.enable_cache
    page_args["style"] = _make_dictionary(page.style, proj_ident)
    page_dict["CSS"] = page_args
    return page_dict


def _create_jsonpage(page_dict, page, ident, proj_ident):
    "Creates a dictionary for the page"
    page_args = OrderedDict()
    page_args["enable_cache"] = page.enable_cache
    page_args["content"] = OrderedDict()
    if page.content:
        page_args["content"] = _make_dictionary(page.content, proj_ident)
    page_dict["JSON"] = page_args
    return page_dict


def _create_svgpage(page_dict, page, ident, proj_ident):
    "Creates a dictionary for the page"
    page_args = OrderedDict()
    page_args["enable_cache"] = page.enable_cache
    page_args["width"] = str(page.width)
    page_args["height"] = str(page.height)
    if page.css_list:
        page_args["css_list"] = _make_list(page.css_list, proj_ident)
    page_dict["SVG"] = page_args
    page_dict["svg"] = page.svg.outline(proj_ident)

    return page_dict


def _create_templatepage(page_dict, page, ident, proj_ident):
    "Creates a dictionary for the page"
    page_args = OrderedDict()

    page_args["show_backcol"] = page.show_backcol
    page_args["last_scroll"] = page.last_scroll
    page_args["interval"] = page.interval

    # convert page.interval_target to ident, label or None
    interval_target = skiboot.make_ident_or_label(page.interval_target, proj_ident)
    if isinstance(interval_target, skiboot.Ident):
        if interval_target.proj == proj_ident:
            page_args["interval_target"] = interval_target.num
        else:
            # ident is another project, put the full ident
            page_args["interval_target"] = interval_target.to_comma_str()
    else:
        # interval_target is label or None
        page_args["interval_target"] = interval_target


    # convert page.catch_to_html to ident, label or None
    catch_to_html = skiboot.make_ident_or_label(page.catch_to_html, proj_ident)
    if isinstance(catch_to_html, skiboot.Ident):
        if catch_to_html.proj == proj_ident:
            page_args["catch_to_html"] = catch_to_html.num
        else:
            # ident is another project, put the full ident
            page_args["catch_to_html"] = catch_to_html.to_comma_str()
    else:
        # catch_to_html is label or None
        page_args["catch_to_html"] = catch_to_html

    page_args["lang"] = page.lang
    if page.backcol:
        page_args["backcol"] = page.backcol

    if page.default_error_widget:
        page_args["default_error_widget"] = page.default_error_widget.sw_tuple()

    page_dict["TemplatePage"] = page_args
    page_dict["head"] = page.head.outline(proj_ident)
    page_dict["body"] = page.body.outline(proj_ident)

    return page_dict





#
# This script is meant to be run from the command line using
#
# python3 -m skipole mynewproj /path/to/projectfiles
#
#

import sys, os, re, shutil

from . import version

DESCRIPTION = """Usage is

python3 -m skipole mynewproj /path/to/projectfiles

Which creates a directory /path/to/projectfiles
containing a sub directy and a python file:

mynewproj.py - a minimal python file which you willdevelop further.

mynewproj - a subdirectory containing data and static files for your new project

You should replace 'mynewproj' with your preferred name for a new project.

The path "/path/to/projectfiles" must be given, and is the path to a directory
where you will develop your project. Multiple projects can be created in one
'projectfiles' directory, or you could have multiple such directories holding
different projects.

If mynewproj already exists in the directory, it will not be changed.
"""


# a search for anything none-alphanumeric and not an underscore
_AN = re.compile('[^\w]')


args = sys.argv

if len(args) == 3:
    project_name = args[1]
    if args[2].startswith('-'):
        print("Invalid filepath. " + DESCRIPTION)
        sys.exit(3)
    projectfiles = os.path.abspath(os.path.expanduser(args[2]))
    if _AN.search(project_name):
        print( "Error: Invalid project name, alphanumeric only")
        sys.exit(1)
    if '_' in project_name:
        print( "Error: Invalid project name, alphanumeric only (no underscore).")
        sys.exit(1)
    if (project_name == 'skis') or (project_name=='skiadmin'):
        print("Error: This project name is reserved")
        sys.exit(2)
elif len(args) == 2:
    if args[1] == "--version":
        print(version)
        sys.exit(0)
    elif (args[1] == "-h") or (args[1] == "--help"):
        print(DESCRIPTION)
        sys.exit(0)
    else:
        print("Unrecognised option. " + DESCRIPTION)
        sys.exit(3)
else:
    print( "Invalid input. " + DESCRIPTION)
    sys.exit(3)


# get the location of the directories to be copied
template_directory = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'templates')

template_newproj_directory = os.path.join(template_directory, 'newproj')
if not os.path.isdir(template_newproj_directory):
    print("Error: Cannot find the template data for the newproj project")
    sys.exit(7)
template_newproj_pyfile = os.path.join(template_directory, 'newproj.py')
if not os.path.isfile(template_newproj_pyfile):
    print("Error: Cannot find the template data for the newproj project")
    sys.exit(8)


project_directory = os.path.join(projectfiles, project_name)
project_pyfile = os.path.join(projectfiles, project_name + ".py")

# newproj python file
newproj_pyfile = os.path.join(projectfiles, 'newproj.py')


# If the given projectfiles directory does not exist
if not os.path.isdir(projectfiles):

    # create the directory by copying the template_directory
    shutil.copytree(template_directory, projectfiles)

    # new directory which has been created
    newproj_directory = os.path.join(projectfiles, 'newproj')

    # change the name of the newproj_directory to project_name
    os.rename(newproj_directory, project_directory)
else:
    # The given projectfiles does exist, so use it for the new project

    if os.path.isdir(project_directory) or os.path.isfile(project_pyfile):
        print("Project %s already exists, and has not been altered." % (project_name,))
        sys.exit(0)

    # create a new project
    shutil.copytree(template_newproj_directory, project_directory)
    shutil.copyfile(template_newproj_pyfile, newproj_pyfile)


# new_code_path -> project_pyfile
# newproj_file -> newproj_pyfile


print("Creating %s" % (project_pyfile,))

# change newproj.py to project_name.py, including the contents
## READ newproj.py

with open(newproj_pyfile, "r") as tnf:
    templatecontents = tnf.read()
os.remove(newproj_pyfile)
## REPLACE newproj with the new project name
newcontents = templatecontents.replace("newproj", project_name)
## REPLACE the 'os.path.dirname(os.path.realpath(__file__))' call with the new projectfiles
newcontents = newcontents.replace("os.path.dirname(os.path.realpath(__file__))", "\"" + projectfiles + "\"")
## WRITE the new file
with open(project_pyfile, "w") as pf:
    pf.write(newcontents)

# change data/project.json contents
## READ project.json
data_file = os.path.join(project_directory, "data", "project.json")
with open(data_file, "r") as tdf:
    templatedatacontents = tdf.read()
os.remove(data_file)
## REPLACE newproj with the new project name
newdatacontents = templatedatacontents.replace("newproj", project_name)
## WRITE the new file
with open(data_file, "w") as tdf2:
    tdf2.write(newdatacontents)

print("""Use the command:
python3 %s
To serve the project at url localhost:8000
and localhost:8000/skiadmin for the admin pages""" % (project_pyfile,))
sys.exit(0)


